"""
Tries to figure out the file type based on content.
"""
import csv
import re


def magic(file_path):
    """
    Based on the first few blocks of `file_path` contents, figure out the likely
    data file serialization type.
    """
    # if XML, it's likely rdf/xml
    # if JSON, it's likely json-ld
    # we go through the various text-based formats before finally seeing if we can parse it
    # as CSV or similar (regular separator character, line feeds, etc.)
    with open(file_path, "r") as file:
        blocks = file.read(8096)  # read 8k - should be enough to figure out what we have
        if re.match(r'^\s*<\?xml', blocks):
            return 'application/rdf+xml'
        if re.match(r'^\s*<!--', blocks):
            return 'application/rdf+xml'
        if re.match(r'^\s*<[a-zA-Z]+://', blocks):
            return 'application/n-triples'
        if re.match(r'^\s*<\S', blocks):
            return 'application/rdf+xml'
        if re.match(r'^\s*@(base|prefix)\s+', blocks):
            return 'text/turtle'
        if re.match(r'^\s*(BASE|PREFIX)\s+', blocks):
            return 'text/turtle'
        if re.match(r'^\s*[{[]', blocks):
            return 'application/ld+json'
        if _maybe_csv(blocks):
            return 'text/csv'
        return 'text/plain'


def _maybe_csv(blocks):
    """
    Try to figure out if the file is likely to be CSV or similar.
    """
    dialect = csv.Sniffer().sniff(blocks)
    return dialect is not None



import sys

from cement import Controller, ex

from ..core.services import client


class Show(Controller):
    """
    Provides the `ursactl show ...` set of commands.
    """
    class Meta:
        label = 'show'
        stacked_on = 'base'
        stacked_type = 'nested'
        help = 'show something'

    @ex(help='show platform endpoint')
    def platform(self):
        print(self.app.config.get('ursactl', 'platform'))

    @ex(help='show default project')
    def project(self):
        print(self.app.config.get('ursactl', 'project'))

    @ex(help='show token')
    def token(self):
        iam_client = client('iam', self.app)
        token = iam_client.get_token()
        if token:
            print(token)
        else:
            sys.exit(1)



from ursactl.core.services import client
from ursactl.core._base import Base
from .running_agent import RunningAgent

import json


class Agent(Base):
    """
    Provides access to an agent definition.
    """
    @property
    def client(self):
        if self._client is None:
            self._client = client('planning', self.app)
        return self._client

    @property
    def name(self):
        return self._data['name']

    @property
    def packages(self):
        return self._data['packages']

    @property
    def behaviors(self):
        return self._data['behaviors']

    def run(self, state=None, predicates=None):
        result = self.client.run_agent(
            self.uuid,
            initial_state=json.dumps(state),
            initial_predicates=json.dumps(predicates),
            project_uuid=self.project.uuid
        )

        if result['errors']:
            return None
        running_agent_id = result['result']['id']
        return RunningAgent(uuid=running_agent_id, app=self.app, client=self.client, project=self.project)

    @property
    def _data(self):
        if self._cached_data is None:
            if self.uuid is None:
                self._cached_data = {
                    'name': None,
                    'packages': None,
                    'behaviors': None,
                    'project_uuid': None
                }
            else:
                self._cached_data = self.client.get_agent(self.uuid)
        return self._cached_data



from ursactl.core.services import client
from ursactl.core._base import Base


class RunningAgent(Base):
    """
    Provides access to a running agent.
    """
    @property
    def client(self):
        if self._client is None:
            self._client = client('planning', self.app)
        return self._client

    def send_event(self, domain, name, params):
        result = self.client.send_event_to_agent(
            self.uuid,
            {
                'domain': domain,
                'name': name,
                'params': params
            }
        )
        return result['result']

    def send_events(self, events):
        actions = []
        for event in events:
            result = self.client.send_event_to_agent(self.uuid, event)
            if result['result']:
                actions.extend(result['result'])
        return actions



def get_provider_info():
    return {
        'package-name': 'ursactl',
        'name': 'Ursa Frontier',
        'description': 'Provides access to Ursa Frontier SaaS Platform datasets and data generation pipelines.',
        'connection-types': [
            {
                'connect-type': 'ursa_frontier_platform',
                'hook-class-name': 'ursactl.airflow_provider.hooks.UrsaCtlHook'
            }
        ]
    }




from cement import Controller
from cement.utils.version import get_version_banner
from ..core.version import get_version

VERSION_BANNER = """
Ursa Frontier CLI. %s
%s
""" % (get_version(), get_version_banner())


class Base(Controller):
    class Meta:
        label = 'base'

        title = 'verbs'

        # text displayed at the top of --help output
        description = 'Command line control script and library for Ursa Frontier SaaS Platform.'

        # text displayed at the bottom of --help output
        epilog = 'Usage: ursactl verb ...'

        # controller level arguments. ex: 'ursactl --version'
        arguments = [
            # add a version banner
            (['-v', '--version'],
             {'action': 'version',
                'version': VERSION_BANNER}),
            (['-o', '--output'],
             {'help': 'Select an alternate output format (json, yaml)',
              'action': 'store',
              'dest': 'output'}),
        ]

    def _default(self):
        """Default action if no sub-command is passed."""

        self.app.args.print_help()



"""
Manages service discovery.
"""
from typing import Optional

from .ape_client import ApeClient
from .iam_client import IamClient
from .planning_client import PlanningClient
from .project_client import ProjectClient
from .dss_client import DssClient

SERVICE_CLIENTS = {
    'ape': ApeClient,
    'dss': DssClient,
    'iam': IamClient,
    'planning': PlanningClient,
    'projects': ProjectClient
}


def client(service: Optional[str] = None, app=None):
    """
    Returns a client appropriate for the service. The optional `app` parameter
    allows the reuse of an UrsaCtl object to manage configuration.
    """
    if app is None:
        from ursactl.main import UrsaCtl
        app = UrsaCtl()
        app.reload()

    platform = app.config.get('ursactl', 'platform')

    if service in SERVICE_CLIENTS:
        return SERVICE_CLIENTS[service](platform, app)



"""
IAM API client.
"""
import sys

import requests
from requests.auth import HTTPBasicAuth
from tinydb import Query
import configparser


class IamClient:
    """
    A client for accessing IAM functions in the Ursa Frontier platform.

    This is limited to using API credentials to get a token that can be
    used with other platform services.
    """
    def __init__(self, endpoint: str, app):
        self.app = app
        self.endpoint = endpoint

    def clear_token(self):
        """
        Used to remove a cached token when it doesn't work.
        """
        try:
            api_key = self.app.config.get('ursactl', 'api_key')
        except configparser.NoOptionError:
            print("You need to set the api_key and api_secret in the ursactl.conf file.")
            sys.exit(1)

        query = Query()
        self.app.db.remove(query.api_key == api_key)

    def get_token(self):
        """
        Retrieves a token that can be used with other services.
        """
        try:
            api_key = self.app.config.get('ursactl', 'api_key')
            api_secret = self.app.config.get('ursactl', 'api_secret')
        except configparser.NoOptionError:
            print("You need to set the api_key and api_secret in the ursactl.conf file.")
            sys.exit(1)

        # check if it's in the tinydb -- if so, return it
        query = Query()
        results = self.app.db.search(query.api_key == api_key)

        if results:
            return results[0]['token']

        # otherwise, get a token, save it to tinydb, and then return it
        response = requests.get(
            f"{self.endpoint}api/token",
            auth=HTTPBasicAuth(api_key, api_secret))

        if response.status_code != 200:
            return

        token = response.json()["token"]
        self.app.db.insert({'api_key': api_key, 'token': token})
        return token




from ursactl.main import UrsaCtlTest


def test_ursactl():
    # test ursactl without any subcommands or arguments
    with UrsaCtlTest() as app:
        app.run()
        assert app.exit_code == 0


def test_ursactl_debug():
    # test that debug mode is functional
    argv = ['--debug']
    with UrsaCtlTest(argv=argv) as app:
        app.run()
        assert app.debug is True



"""
Base class for API clients.
"""

import re
import requests

from python_graphql_client import GraphqlClient

from ursactl.core.exc import UrsaNotAuthorized


class Base:
    """
    Base class for clients.
    """
    def __init__(self, endpoint: str, app):
        from .. import services

        self.app = app
        self.endpoint = endpoint
        self.iam_client = services.client('iam', app)
        self.client = GraphqlClient(endpoint=f"{endpoint}api")

    def raw_query(self, query, variables={}, reauthorize=True):
        try:
            result = self.client.execute(query=query, variables=variables, headers={
                "authorization": f"Bearer {self.iam_client.get_token()}"
            })

            if 'errors' in result:
                for error in result['errors']:
                    if error['code'] == 'Forbidden':
                        if reauthorize:
                            self.iam_client.clear_token()
                            return self.raw_query(query, variables, reauthorize=False)
                        raise UrsaNotAuthorized("Not authorized")
                    else:
                        print(error['message'])
        except requests.exceptions.HTTPError as e:
            if e.response.status_code == 401:
                if reauthorize:
                    self.iam_client.clear_token()
                    return self.raw_query(query, variables, reauthorize=False)
                raise UrsaNotAuthorized("Not authorized")
            else:
                raise e
        return result

    @staticmethod
    def is_uuid(s: str) -> bool:
        """
        Test if the string looks like a UUID.
        """
        if s is None:
            return False
        return re.match(r'^[0-9a-f]{8}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{12}$', s)



"""
DataStore Service client
"""

import requests
import urllib.parse

from ursactl.core.exc import UrsaNotAuthorized

from ._base import Base

UPDATE_DATASET_MUTATION = """\
mutation update_dataset($id: ID!, $description: string!) {
    updateDataset(id: $id, description: $description) {
        errors { message }
    }
}
"""

GET_DATASET_PATH_QUERY = """\
query get_dataset($id: ID!) {
    dataset(id: $id) {
        path
        project { id name user { handle } }
    }
}
"""

GET_DATASET_QUERY = """\
query get_dataset($id: ID!) {
    dataset(id: $id) {
        id
        path
        contentType
        description
        size
        project { id }
    }
}
"""

GET_DATASET_BY_PATH_QUERY = """\
query get_dataset($projectId: String!, $path: String!) {
    project(handleName: $projectId) {
        datasets(filter: { path: { eq: $path } }, limit: 1) {
            id
            path
            contentType
            description
            size
            project { id }
        }
    }
}
"""

DELETE_DATASET_MUTATION = """\
mutation delete_dataset($id: ID!) {
    deleteDataset(id: $id) {
        errors {
            message
        }
    }
}
"""

LIST_DATASETS_FOR_PROJECT_QUERY = """\
query listDatasets($projectId: String!) {
    project(handleName: $projectId) {
        datasets(limit:100) {
            id
            path
            size
            content_type
        }
    }
}
"""


class DssClient(Base):
    @staticmethod
    def _raise_if_none(value, var_name):
        if value is None:
            raise ValueError(f"{var_name} must be provided")

    def create_dataset(self, project_uuid=None, path=None, description=None, data=None, data_type=None, reauthorize=True):
        """
        Creates a dataset, returning the uuid and id of the dataset.

        If no data is provided, the dataset is created with no attached data.
        Data can be attached later.
        """
        # we first use the graphql to create the dataset
        # then we use a PUT to upload the data
        # the Python graphql client doesn't support multi-part HTTP requests
        self._raise_if_none(project_uuid, 'project_uuid')
        self._raise_if_none(path, 'path')
        self._raise_if_none(data, 'data')
        self._raise_if_none(data_type, 'data_type')

        # first, we POST to the right URL to upload the contents of the file
        with open(data, 'rb') as file:
            contents = file.read()
        res = requests.post(
            url=f"{self.endpoint}api/dss/{project_uuid}/{path}",
            data=contents,
            headers={
                'Content-Type': data_type,
                'Authorization': f"Bearer {self.iam_client.get_token()}"
            })

        if res.status_code == 401:  # uh oh!
            if reauthorize:
                self.iam_client.clear_token()
                return self.create_dataset(
                    project_uuid=project_uuid,
                    path=path,
                    description=description,
                    data=data,
                    data_type=data_type,
                    reauthorize=False)
            raise UrsaNotAuthorized("forbidden")
        if res.status_code == 409:
            return {'accepted': False}
        if res.status_code >= 500:
            print("Unable to create dataset. Internal platform error.")
            return {'accepted': False}

        dataset_id = res.headers['x-dataset-id']
        # if successful, we get the dataset id from the response headers
        # if we have a description, we use GraphQL to set the description

        if description is not None:
            variables = {
                'id': dataset_id,
                'description': description
            }

            query_response = self.raw_query(query=UPDATE_DATASET_MUTATION, variables=variables)
            if 'errors' in query_response:
                return {'accepted': False}

        return {
            'accepted': True,
            'id': dataset_id
        }

    def get_dataset_url(self, project_uuid=None, dataset=None):
        if self.is_uuid(dataset):
            variables = {
                'id': dataset
            }
            query_response = self.raw_query(query=GET_DATASET_PATH_QUERY, variables=variables)
            if 'errors' in query_response:
                return None
            path = query_response['data']['dataset']['path']
            # handle_name = query_response['data']['dataset']['project']['handleName']
            owner_handle = query_response['data']['dataset']['project']['user']['handle']
            project_name = query_response['data']['dataset']['project']['name']
            handle_name = f"{owner_handle}/{project_name}"
            return f"{self.endpoint}api/dss/{urllib.parse.quote(handle_name)}/{urllib.parse.quote(path)}"
        else:
            return f"{self.endpoint}api/dss/{urllib.parse.quote(project_uuid)}/{urllib.parse.quote(dataset)}"

    def get_dataset_details(self, dataset_id, project_uuid=None):
        if self.is_uuid(dataset_id):
            variables = {
                'id': dataset_id
            }
        else:
            dataset = self.get_dataset(dataset_id, project_uuid=project_uuid)
            if dataset is None:
                return None
            variables = {
                'id': dataset['id']
            }
        query_response = self.raw_query(query=GET_DATASET_QUERY, variables=variables)
        if 'errors' in query_response:
            return None
        return query_response['data']['dataset']

    def get_dataset(self, dataset_id, project_uuid=None):
        if self.is_uuid(dataset_id):
            variables = {
                'id': dataset_id
            }
            query_response = self.raw_query(query=GET_DATASET_QUERY, variables=variables)
            return query_response['data']['dataset']
        else:
            variables = {
                'projectId': project_uuid,
                'path': dataset_id
            }
            query_response = self.raw_query(query=GET_DATASET_BY_PATH_QUERY, variables=variables)
            return query_response['data']['project']['datasets'][0]

    def download_dataset(self, url, local_path, reauthorize=True):
        with open(local_path, 'wb') as fd:
            r = requests.get(url, stream=True, headers={
                    "authorization": f"Bearer {self.iam_client.get_token()}"
                })
            if r.status_code >= 400:
                if reauthorize:
                    self.iam_client.clear_token()
                    return self.download_dataset(url, local_path, reauthorize=False)
                return False
            for chunk in r.iter_content(chunk_size=1024*1024):
                fd.write(chunk)
        return True

    def download_dataset_content(self, url, reauthorize=True):
        local = bytearray()
        r = requests.get(url, stream=True, headers={
                "authorization": f"Bearer {self.iam_client.get_token()}"
            })
        if r.status_code >= 400:
            if reauthorize:
                self.iam_client.clear_token()
                return self.download_dataset_content(url, reauthorize=False)
            return False
        for chunk in r.iter_content(chunk_size=1024*1024):
            local += chunk
        content = local.decode('utf-8')
        return content

    def delete_dataset(self, dataset_id, project_uuid=None):
        if self.is_uuid(dataset_id):
            variables = {
                'id': dataset_id
            }
        else:
            dataset = self.get_dataset(dataset_id, project_uuid=project_uuid)
            if dataset is None:
                return False
            variables = {
                'id': dataset['id']
            }
        query_response = self.raw_query(query=DELETE_DATASET_MUTATION, variables=variables)
        if 'errors' in query_response and query_response['errors']:
            return False
        return True

    def list_datasets(self, project_scope=None):
        if project_scope is None:
            raise ValueError("project_scope must be provided")
        variables = {}
        query = LIST_DATASETS_FOR_PROJECT_QUERY
        variables = {
            'projectId': project_scope
        }
        query_response = self.raw_query(query=query, variables=variables)

        if 'errors' in query_response:
            return []
        return query_response['data']['project']['datasets']



from ursactl.core.services import client
from ursactl.core._base import Base


class Project(Base):
    """
    Provides access to a project and its related resources.
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if self._uuid is None:
            self._uuid = self.app.config.get('ursactl', 'project')

    @property
    def client(self):
        if self._client is None:
            self._client = client('usage', self.app)
        return self._client

    @property
    def name(self):
        return self._data['name']

    @property
    def description(self):
        return self._data['description']

    @property
    def is_archived(self):
        return self._data['isArchived']

    def Agent(self, *args, **kwargs):
        from ursactl.core.agent import Agent as AgentClass

        return AgentClass(*args, project_uuid=self.uuid, **kwargs)

    def agents(self):
        """
        Returns a generator listing all agents belonging to the project.
        """
        from ursactl.core.agent import Agent

        if self.uuid is None:
            return []
        planning_client = client('planning', self.app)
        return (
            Agent(client=planning_client, app=self.app, project=self, **info)
            for info in planning_client.list_agents(project_scope=self.uuid)
        )

    def agent(self, name):
        """
        Returns an agent with the given name.
        """
        planning_client = client('planning', self.app)
        info = planning_client.get_agent(name, project_uuid=self.uuid)
        return self.Agent(uuid=info['id'], project=self, client=planning_client, app=self.app, **info)

    def Dataset(self, *args, **kwargs):
        from ursactl.core.dataset import Dataset as DatasetClass

        return DatasetClass(*args, project=self, app=self.app, **kwargs)

    def datasets(self):
        """
        Returns a generator listing all datasets belonging to the project.
        """
        from ursactl.core.dataset import Dataset

        if self.uuid is None:
            return []
        dss_client = client('dss', self.app)
        return (
            Dataset(client=dss_client, app=self.app, project=self, **info)
            for info in dss_client.list_datasets(project_scope=self.uuid)
        )

    def Pipeline(self, *args, **kwargs):
        from ursactl.core.pipeline import Pipeline as PipelineClass

        return PipelineClass(*args, project=self, **kwargs)

    def pipelines(self):
        """
        Returns a generator listing all pipelines belonging to the project.
        """
        from ursactl.core.pipeline import Pipeline

        if self.uuid is None:
            return []
        ape_client = client('ape', self.app)
        return (
            Pipeline(client=ape_client, app=self.app, project=self, **info)
            for info in ape_client.list_pipelines(project_scope=self.uuid)
        )

    def pipeline(self, path):
        """
        Returns a pipeline with the given path.
        """
        ape_client = client('ape', self.app)
        info = ape_client.get_pipeline(path=path, project_uuid=self.uuid)
        return self.Pipeline(uuid=info['id'], client=ape_client, app=self.app, **info)

    @property
    def _data(self):
        if self._cached_data is None:
            if self.uuid is None:
                self._cached_data = {
                    'name': None,
                    'description': None,
                    'isArchived': None
                }
            else:
                self._cached_data = self.client.get_project_details(self.uuid)
        return self._cached_data



"""
Project API client.
"""
from typing import Optional

from ._base import Base

LIST_PROJECTS_QUERY = """\
{
    projects {
        id
        name
        description
        isArchived
        user { handle }
    }
}
"""

CREATE_PROJECT_MUTATION = """\
mutation create_project($input: CreateProjectInput!) {
    createProject(input: $input) {
        result {
            id
        }
        errors {
            message
        }
    }
}
"""

UPDATE_PROJECT_MUTATION = """\
mutation update_project($id: ID!, $input: UpdateProjectInput!) {
    updateProject(id: $id, input: $input) {
        errors {
            message
        }
    }
}
"""

CLOSE_PROJECT_MUTATION = """\
mutation archive_project($id: ID!) {
    archiveProject(id: $id) {
        result {
            isArchived
        }
        errors {
            message
        }
    }
}
"""

OPEN_PROJECT_MUTATION = """\
mutation unarchive_project($id: ID!) {
    unarchiveProject(id: $id) {
        result {
            isArchived
        }
        errors {
            message
        }
    }
}
"""


class ProjectClient(Base):

    def list(self):
        """
        Lists projects for the authenticated user.
        """

        data = self.raw_query(query=LIST_PROJECTS_QUERY)
        return data['data']['projects']

    def create(self, name: str, description: Optional[str] = None):
        """
        Create a new project.
        """
        variables = {
            'input': {
                'name': name,
                'description': description
            }
        }
        data = self.raw_query(query=CREATE_PROJECT_MUTATION, variables=variables)
        return data['data']['createProject']['result']

    def update(self, id: str, name: Optional[str] = None, description: Optional[str] = None):
        """
        Update an existing project.
        """
        variables = {
            'id': id,
            'input': {
            }
        }

        if name is not None:
            variables['input']['name'] = name
        if description is not None:
            variables['input']['description'] = description

        data = self.raw_query(query=UPDATE_PROJECT_MUTATION, variables=variables)
        result = data['data']['updateProject']['errors']
        if result:
            result = {
                'errors': result,
                'accepted': False
            }
        else:
            result = {}
            result['accepted'] = True
        return result

    def unarchive(self, id: str):
        """
        Unarchives an archived project.
        """
        data = self.raw_query(query=OPEN_PROJECT_MUTATION, variables={'id': id})
        return data['data']['unarchiveProject']

    def archive(self, id: str):
        """
        Archive a project.
        """
        data = self.raw_query(query=CLOSE_PROJECT_MUTATION, variables={'id': id})
        return data['data']['archiveProject']



"""
Exception classes for Ursactl
"""


class UrsaCtlError(Exception):
    """Generic errors."""
    pass


class UrsaNotAuthorized(UrsaCtlError):
    """Authentication/authorization errors."""
    pass


class UrsaProjectNotDefined(UrsaCtlError):
    """Project required but not defined."""
    pass




from cement.utils.version import get_version as cement_get_version

VERSION = (0, 3, 2)
PRERELEASE = ('final', 0)


def get_version(version=VERSION):
    return cement_get_version(version + PRERELEASE)



"""
Planning capabilities client.
"""

from ._base import Base
import json

LIST_PACKAGES_QUERY = """\
query list_packages($id: String!) {
    project(handleName: $id) {
        packages {
            id
            name
        }
    }
}
"""

LIST_AGENTS_QUERY = """\
query list_packages($id: String!) {
    project(handleName: $id) {
        agents {
            id
            name
        }
    }
}
"""

CREATE_PACKAGE_MUTATION = """\
mutation create_package($input: CreatePackageInput!) {
    createPackage(input: $input) {
        errors {
            code
            message
            fields
        }
        result {
            id
            name
        }
    }
}
"""

UPDATE_PACKAGE_MUTATION = """\
mutation update_package($input: UpdatePackageInput!) {
    updatePackage(input: $input) {
        errors {
            code
            message
            fields
        }
        result {
            id
            name
        }
    }
}
"""

GET_PACKAGE_BY_NAME = """\
query get_package($name: String!, $project_id: String!) {
    project(handleName: $project_id) {
        packages(limit: 1, filter: { name: { eq: $name } }) {
            id
            name
        }
    }
}
"""

GET_PACKAGE_BY_ID = """\
query get_package($id: ID!) {
    package(id: $id) {
        id
        name
    }
}
"""

LIST_DOMAINS_BY_PACKAGE_NAME = """\
query get_package($name: String!, $project_id: String!) {
    project(handleName: $project_id) {
        packages(limit: 1, filter: { name: { eq: $name } }) {
            id
            name
            domains {
                id
                name
            }
        }
    }
}
"""

LIST_DOMAINS_BY_PACKAGE_ID = """\
query get_package($package_id: ID!) {
    package(id: $package_id) {
        id
        name
        domains {
            id
            name
        }
    }
}
"""

DELETE_PACKAGE_MUTATION = """\
mutation delete_package($id: ID!) {
    deletePackage(id: $id) {
        errors {
            code
            message
            fields
        }
        result {
            id
            name
        }
    }
}
"""

CREATE_AGENT_MUTATION = """\
mutation create_agent($input: CreateAgentInput!) {
    createAgent(input: $input) {
        errors {
            code
            message
            fields
        }
        result {
            id
            name
        }
    }
}
"""

DELETE_AGENT_MUTATION = """\
mutation delete_agent($id: ID!) {
    deleteAgent(id: $id) {
        errors {
            code
            message
            fields
        }
        result {
            id
            name
        }
    }
}
"""

GET_AGENT_BY_NAME = """\
query get_agent($name: String!, $project_id: String!) {
    project(handleName: $project_id) {
        agents(limit: 1, filter: { name: { eq: $name } }) {
            id
            name
        }
    }
}
"""

GET_AGENT_BY_ID = """\
query get_agent($id: ID!) {
    agent(id: $id) {
        id
        name
    }
}
"""

RUN_AGENT_MUTATION = """\
mutation run_agent($input: RunAgentInput!) {
    runAgent(input: $input) {
        errors {
            code
            message
            fields
        }
        result {
            id
        }
    }
}
"""

STOP_AGENT_MUTATION = """\
mutation stop_agent($id: ID!) {
    stopAgent(id: $id) {
        errors {
            code
            message
            fields
        }
        result {
            id
        }
    }
}
"""

LIST_RUNNING_AGENTS_BY_PROJECT = """\
query list_agents_by_project($id: String!) {
    project(handleName: $id) {
        agents {
            name
            runningAgents {
                id
                startedAt
                status
            }
        }
    }
}
"""

LIST_RUNNING_AGENTS_BY_AGENT = """\
query list_agents_by_agent($id: ID!) {
    agent(id: $id) {
        name
        runningAgents {
            id
            startedAt
            status
        }
    }
}
"""

SEND_EVENT_TO_AGENT_MUTATION = """\
mutation send_event($agentId: ID!, $domain:String!, $name:String!, $params:Json!) {
  sendEventToAgent(agentId:$agentId,domain:$domain,name:$name,params:$params) {
    errors {
        code
        message
        fields
    }
    result {
        domain
        name
        params
    }
  }
}
"""


class PlanningClient(Base):
    """-
    Client providing access to the planning capabilities.
    """
    def list_packages(self, project_uuid=None):
        """
        Return a list of available packages.
        """

        query_response = self.raw_query(query=LIST_PACKAGES_QUERY, variables={
            'id': project_uuid
        })
        return query_response['data']['project']['packages']

    def list_agents(self, project_uuid=None):
        """
        Return a list of available agents.
        """
        query_response = self.raw_query(query=LIST_AGENTS_QUERY, variables={
            'id': project_uuid
        })
        return query_response['data']['project']['agents']

    def list_running_agents(self, agent_id=None, project_uuid=None):
        """
        Return a list of running agents.
        """
        if agent_id is not None:
            if self.is_uuid(agent_id):
                variables = {'id': agent_id}
            else:
                agent = self.get_agent(agent_id, project_uuid)
                if agent is None:
                    return []
                variables = {'id': agent['id']}
            query_response = self.raw_query(query=LIST_RUNNING_AGENTS_BY_AGENT, variables=variables)
            agent_name = query_response['data']['agent']['name']
            return [{'agent': agent_name, **info} for info in query_response['data']['agent']['runningAgents']]

        elif project_uuid is not None:
            query_response = self.raw_query(query=LIST_RUNNING_AGENTS_BY_PROJECT, variables={
                'id': project_uuid
            })
            return [
                {'agent': agent['name'], **info}
                for agent in query_response['data']['project']['agents']
                for info in agent['runningAgents']
            ]

    def create_package(self, project_uuid=None, content=None):
        """
        Create a new package.
        """
        mutation_response = self.raw_query(query=CREATE_PACKAGE_MUTATION, variables={
            'input': {
                'source': content,
                'project': project_uuid
            }
        })
        return mutation_response['data']['createPackage']

    def update_package(self, project_uuid=None, content=None):
        """
        Update an existing package.
        """
        mutation_response = self.raw_query(query=UPDATE_PACKAGE_MUTATION, variables={
            'input': {
                'source': content,
                'project': project_uuid
            }
        })
        return mutation_response['data']['updatePackage']

    def create_agent(self, project_uuid=None, name=None, packages=None, behaviors=None):
        """
        Create a new agent.
        """
        mutation_response = self.raw_query(query=CREATE_AGENT_MUTATION, variables={
            'input': {
                'name': name,
                'packages': packages,
                'behaviors': behaviors,
                'project': project_uuid
            }
        })
        return mutation_response['data']['createAgent']

    def get_package(self, name=None, project_uuid=None):
        """
        Get the id and basic info about a package.
        """
        if self.is_uuid(name):
            variables = {
                'id': name
            }
            query = GET_PACKAGE_BY_ID
        else:
            variables = {
                'name': name,
                'project_id': project_uuid
            }
            query = GET_PACKAGE_BY_NAME
        result = self.raw_query(query=query, variables=variables)
        if query == GET_PACKAGE_BY_ID:
            return result['data']['package']
        else:
            return result['data']['project']['packages'][0]

    def get_agent(self, name=None, project_uuid=None):
        """
        Get the id and basic info about an agent.
        """
        if self.is_uuid(name):
            variables = {
                'id': name
            }
            query = GET_AGENT_BY_ID
        else:
            variables = {
                'name': name,
                'project_id': project_uuid
            }
            query = GET_AGENT_BY_NAME
        result = self.raw_query(query=query, variables=variables)
        if query == GET_AGENT_BY_ID:
            return result['data']['agent']
        else:
            return result['data']['project']['agents'][0]

    def delete_package(self, id, project_uuid=None):
        """
        Deletes a package.
        """
        if self.is_uuid(id):
            variables = {
                'id': id
            }
        else:
            package = self.get_package(name=id, project_uuid=project_uuid)
            if package is None:
                return {}
            variables = {'id': package['id']}

        result = self.raw_query(query=DELETE_PACKAGE_MUTATION, variables=variables)

        return result['data']['deletePackage']

    def delete_agent(self, id, project_uuid=None):
        """
        Deletes an agent.
        """
        if self.is_uuid(id):
            variables = {
                'id': id
            }
        else:
            agent = self.get_agent(name=id, project_uuid=project_uuid)
            if agent is None:
                return {}
            variables = {'id': agent['id']}

        result = self.raw_query(query=DELETE_AGENT_MUTATION, variables=variables)

        return result['data']['deleteAgent']

    def run_agent(self, id, project_uuid=None, initial_predicates=None, initial_state=None):
        """
        Runs an agent.
        """
        variables = {'input': {'project': project_uuid}}
        if self.is_uuid(id):
            agent = self.get_agent(name=id, project_uuid=project_uuid)
            if agent is None:
                return {}
            variables['input']['agent'] = agent['name']
        else:
            agent = self.get_agent(name=id, project_uuid=project_uuid)
            if agent is None:
                return {}
            variables['input']['agent'] = id

        if initial_state is not None:
            variables['input']['initialState'] = initial_state
        if initial_predicates is not None:
            variables['input']['initialPredicates'] = initial_predicates

        result = self.raw_query(query=RUN_AGENT_MUTATION, variables=variables)

        return result['data']['runAgent']

    def send_event_to_agent(self, agent_instance_id, event):
        """
        Sends an event to an agent.
        """
        variables = {'agentId': agent_instance_id, **event}
        variables['params'] = json.dumps(variables['params'])

        result = self.raw_query(query=SEND_EVENT_TO_AGENT_MUTATION, variables=variables)

        return result['data']['sendEventToAgent']

    def stop_agent(self, id):
        """
        Stops an agent.
        """
        variables = {
            'id': id
        }

        result = self.raw_query(query=STOP_AGENT_MUTATION, variables=variables)

        return result['data']['stopAgent']

    def list_domains(self, id=None, project_uuid=None):
        """
        List domains in a package.
        """

        if self.is_uuid(id):
            variables = {
                'package_id': id
            }
            query = LIST_DOMAINS_BY_PACKAGE_ID
        else:
            variables = {
                'name': id,
                'project_id': project_uuid
            }
            query = LIST_DOMAINS_BY_PACKAGE_NAME

        result = self.raw_query(query=query, variables=variables)

        if query == LIST_DOMAINS_BY_PACKAGE_ID:
            return result['data']['package']['domains']
        else:
            return result['data']['project']['packages'][0]['domains']



"""
Dataset object providing a OOP interface to datasets.
"""
from ursactl.core.services import client
from ursactl.core._base import Base


class Dataset(Base):
    """
    Provides access to a dataset.
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._content = None

    @property
    def client(self):
        if self._client is None:
            self._client = client('dss', self.app)
        return self._client

    def commit(self):
        if self.uuid is None:
            # create rather than update
            # don't support uploading data (yet)
            result = self.client.create_dataset(
                project_uuid=self.project.uuid,
                name=self.name,
                description=self.description
            )
            if result['accepted']:
                self._uuid = result['uuid']
                return True
            return False
        return False

    @property
    def path(self):
        return self._data['path']

    @property
    def description(self):
        return self._data['description']

    @property
    def type(self):
        return self._data['type']

    @property
    def size(self):
        return self._data['size']

    @property
    def url(self):
        return self._data['url']

    @property
    def content(self):
        if self._content is None and self.url is not None:
            self._content = self.client.download_dataset_content(self.url)
        return self._content

    @property
    def created_at(self):
        return self._data['createdAt']

    @property
    def updated_at(self):
        return self._data['updatedAt']

    @property
    def _data(self):
        if self._cached_data is None:
            if self.uuid is None:
                self._cached_data = {
                    'path': None,
                    'description': None,
                    'type': None,
                    'size': None,
                    'url': None,
                    'createdAt': None,
                    'updatedAt': None
                }
            else:
                self._cached_data = self.client.get_dataset_details(self.uuid)
                self._cached_data['url'] = self.client.get_dataset_url(self.project.uuid, self.uuid)

        return self._cached_data



from cement import Controller, ex

from ..core.services import client


class Get(Controller):
    '''
    Provides the 'get' verb.
    '''

    class Meta:
        label = 'get'
        stacked_on = 'base'
        stacked_type = 'nested'
        help = 'get something'

    @ex(help='download dataset',
        arguments=[
            (['--project'], {'help': 'project identifier', 'dest': 'project', 'action': 'store'}),
            (['dataset'], {'help': 'UUID or path of the dataset to download', 'action': 'store'}),
            (['destination'], {'help': 'local file in which the dataset is saved', 'action': 'store'})
        ])
    def dataset(self):
        dataset_uuid = self.app.pargs.dataset
        destination = self.app.pargs.destination
        dss_client = client('dss', self.app)
        url = dss_client.get_dataset_url(project_uuid=self._project_scope, dataset=dataset_uuid)
        if url is None:
            exit(1)
        if not dss_client.download_dataset(url, destination):
            exit(1)

    @property
    def _project_scope(self):
        return self.app.pargs.project or self.app.config.get('ursactl', 'project')



from ursactl.core.services import client
from ursactl.core._base import Base

from .pipeline_run import PipelineRun


class PipelineSweep(Base):
    """
    Provides access to a pipeline sweep.
    """
    def __init__(self, *args, pipeline=None, **kwargs):
        super().__init__(*args, **kwargs)
        self.pipeline = pipeline

    @property
    def client(self):
        if self._client is None:
            self._client = client('ape', self.app)
        return self._client

    @property
    def pipeline_runs(self):
        return [
            PipelineRun(
                uuid=run['id'],
                project=self.project,
                app=self.app,
                client=self.client,
                pipeline=self.pipeline,
                **run) for run in self._data['pipelineRuns']]

    @property
    def job_status(self):
        return self._data['jobStatus']

    def refresh(self):
        if self.uuid is not None:
            if self._cached_data is None:
                self._cached_data = {}
            self._cached_data.update(self.client.get_pipeline_sweep(uuid=self.uuid))

    @property
    def _data(self):
        if self._cached_data is None:
            if self.uuid is None:
                self._cached_data = {
                    'pipelineRuns': [],
                    'jobStatus': None,
                }
            else:
                self._cached_data = self.client.get_pipeline_sweep(uuid=self.uuid)
        return self._cached_data



from pathlib import Path
import sys

from cement import Controller


class Init(Controller):
    """
    Provides the 'init' verb
    """

    class Meta:
        label = 'init'
        stacked_on = 'base'
        stacked_type = 'nested'
        help = 'initialize a directory for use with Ursa Frontier'
        arguments = [
            (['--project'], {
                 'help': 'default project to use with this repository',
                 'dest': 'project',
                 'action': 'store'}),
            (['--dir', '--directory'], {
                 'help': 'directory to initialize (defaults to \'ursa\')',
                 'dest': 'directory',
                 'action': 'store',
                 'default': 'ursa'}),
        ]

    def _default(self):
        root = Path(self.app.pargs.directory)
        if not root.exists():
            root.mkdir(parents=True)
        elif not root.is_dir():
            print("Error: %s exists and is not a directory." % root)
            sys.exit(1)
        with (root / ".ursactl.conf").open(mode='w') as f:
            f.write("### Ursa Frontier Control configuration file\n[ursactl]\n\n")
            if self.app.pargs.project:
                f.write(f"project = {self.app.pargs.project}\n\n")

        for subdir in (
                       'data/transforms',
                       'data/generators',
                       'data/pipelines',
                       'data/datasets',
                       'planning/packages',
                       'planning/agents'
        ):
            (root / subdir).mkdir(parents=True, exist_ok=True)
            (root / subdir / '.keep').touch()



from cement import Controller, ex
import json
import yaml

from ..core.services import client


class Run(Controller):
    """
    Provides the 'run' verb.
    """

    class Meta:
        label = 'run'
        stacked_on = 'base'
        stacked_type = 'nested'
        help = 'run something'

    @ex(help='run agent',
        arguments=[
            (['--project'], {'help': 'project identifier', 'dest': 'project', 'action': 'store'}),
            (['agent'], {'help': 'agent uuid or path', 'action': 'store'})
        ])
    def agent(self):
        planning_client = client('planning', self.app)
        result = planning_client.run_agent(self.app.pargs.agent, project_uuid=self._project_scope)
        self._print(result)

    @ex(help='run pipeline',
        arguments=[
            (['--project'], {'help': 'project identifier', 'dest': 'project', 'action': 'store'}),
            (['--param'], {'help': 'pipeline parameter', 'dest': 'parameters', 'action': 'append'}),
            (['--sweep-param'], {'help': 'pipeline sweep parameter', 'dest': 'sweep_parameters', 'action': 'append'}),
            (['pipeline'], {'help': 'pipeline path', 'action': 'store'})
        ])
    def pipeline(self):
        ape_client = client('ape', self.app)
        pipeline = self.app.pargs.pipeline
        parameters = None
        sweep_parameters = None
        if self.app.pargs.parameters:
            parameters = {
                k: json.loads(v)
                for k, v in dict(item.split('=') for item in self.app.pargs.parameters).items()
            }
        if self.app.pargs.sweep_parameters:
            sweep_parameters = {
                k: json.loads(v)
                for k, v in dict(item.split('=') for item in self.app.pargs.sweep_parameters).items()
            }

        result = ape_client.run_pipeline(
            pipeline,
            parameters=parameters,
            sweep_parameters=sweep_parameters,
            project_uuid=self._project_scope)
        self._print(result)

    def _print(self, data):
        if self.app.pargs.output == 'json':
            print(json.dumps(data))
        else:
            print(yaml.dump(data))

    @property
    def _project_scope(self):
        return self.app.pargs.project or self.app.config.get('ursactl', 'project')





class Base:
    """
    Base class for OOP interface classes.
    """
    def __init__(self, uuid=None, client=None, app=None, project=None, **kwargs):
        self._client = client
        self._app = app
        self._uuid = uuid
        self.project = project

        if any(kwargs):
            self._cached_data = kwargs
        else:
            self._cached_data = None

    @property
    def app(self):
        if self._app is None:
            from ursactl.main import UrsaCtl
            self._app = UrsaCtl()
            self._app.reload()
        return self._app

    @property
    def uuid(self):
        return self._uuid



"""
APE Service client
"""

from json import dumps
import jsonschema

from ._base import Base

LIST_TRANSFORMS_QUERY = """\
query list_transforms($id: String!) {
    project(handleName: $id) {
        transforms {
            id
            name
            path
            type
            isSynced
        }
    }
}
"""

CREATE_TRANSFORM_MUTATION = """\
mutation create_transform($input: CreateTransformInput!) {
    createTransform(input: $input) {
        errors {
            code
            message
            fields
        }
        result {
            id
            name
            path
        }
    }
}
"""

SYNC_TRANSFORM_MUTATION = """\
mutation sync_transform($input: SyncTransformInput!) {
    syncTransform(input: $input) {
        errors {
            code
            message
            fields
        }
        result {
            id
            name
            path
        }
    }
}
"""

UPDATE_TRANSFORM_MUTATION = """\
mutation update_transform($id: ID!, $input: UpdateTransformInput!) {
    updateTransform(id: $id, input: $input) {
        errors {
            code
            message
            fields
        }
        result {
            id
            name
            path
        }
    }
}
"""

GET_TRANSFORM_BY_ID_QUERY = """\
query get_transform($id: ID!) {
    transform(id: $id) {
        path
        description
        configurationSchema
        type
    }
}
"""

GET_TRANSFORM_BY_PATH_QUERY = """\
query get_transform($path: String!, $project: String!) {
    project(handleName: $project) {
        transforms(limit: 1, filter: { path: { eq: $path } }) {
            id
            path
            description
            configurationSchema
            type
        }
    }
}
"""

DELETE_TRANSFORM_MUTATION = """\
mutation delete_transform($id: ID!) {
    deleteTransform(id: $id) {
        errors {
            code
            message
            fields
        }
        result {
            id
            path
        }
    }
}
"""

DELETE_SYNCED_TRANSFORM_MUTATION = """\
mutation delete_synced_transform($id: ID!) {
    deleteSyncedTransform(id: $id) {
        errors {
            code
            message
            fields
        }
        result {
            id
            path
        }
    }
}
"""

CREATE_GENERATOR_MUTATION = """\
mutation create_generator($input: CreateGeneratorInput!) {
    createGenerator(input: $input) {
        errors {
            code
            message
            fields
        }
        result {
            id
            path
        }
    }
}
"""

SYNC_GENERATOR_MUTATION = """\
mutation sync_generator($input: SyncGeneratorInput!) {
    syncGenerator(input: $input) {
        errors {
            code
            message
            fields
        }
        result {
            id
            path
        }
    }
}
"""

UPDATE_GENERATOR_MUTATION = """\
mutation update_generator($id: ID!, $input: UpdateGeneratorInput!) {
    updateGenerator(id: $id, input: $input) {
        errors {
            code
            message
            fields
        }
        result {
            id
            path
        }
    }
}
"""

LIST_GENERATORS_FOR_PROJECT_QUERY = """\
query list_generators($id: String!) {
    project(handleName: $id) {
        generators {
            id
            path
            isSynced
            transform
        }
    }
}
"""

GET_GENERATOR_BY_ID_QUERY = """\
query get_generator($id: ID!) {
    generator(id: $id) {
        path
        project { id }
    }
}
"""

GET_GENERATOR_BY_PATH_QUERY = """\
query get_generator($path: String!, $project_id: String!) {
    project(handleName: $project_id) {
        generators(limit: 1, filter: { path: { eq: $path } }) {
           id
        }
    }
}
"""

DELETE_GENERATOR_MUTATION = """\
mutation delete_generator($id: ID!) {
    deleteGenerator(id: $id) {
        errors {
            code
            message
            fields
        }
        result {
            id
            path
        }
    }
}
"""

DELETE_SYNCED_GENERATOR_MUTATION = """\
mutation delete_synced_generator($id: ID!) {
    deleteSyncedGenerator(id: $id) {
        errors {
            code
            message
            fields
        }
        result {
            id
            path
        }
    }
}
"""

CREATE_PIPELINE_MUTATION = """\
mutation create_pipeline($input: CreatePipelineInput!) {
    createPipeline(input: $input) {
        errors {
            code
            message
            fields
            vars
        }
        result {
            id
            path
        }
    }
}
"""

SYNC_PIPELINE_MUTATION = """\
mutation sync_pipeline($input: SyncPipelineInput!) {
    syncPipeline(input: $input) {
        errors {
            code
            message
            fields
            vars
        }
        result {
            id
            path
        }
    }
}
"""

DELETE_PIPELINE_MUTATION = """\
mutation delete_pipeline($id: ID!) {
    deletePipeline(id: $id) {
        errors {
            code
            message
            fields
        }
        result {
            id
            path
        }
    }
}
"""

DELETE_SYNCED_PIPELINE_MUTATION = """\
mutation delete_synced_pipeline($id: ID!) {
    deleteSyncedPipeline(id: $id) {
        errors {
            code
            message
            fields
        }
        result {
            id
            path
        }
    }
}
"""

UPDATE_PIPELINE_MUTATION = """\
mutation update_pipeline($id: ID!, $input: UpdatePipelineInput!) {
    updatePipeline(id: $id, input: $input) {
        errors {
            code
            message
            fields
        }
        result {
            id
            path
        }
    }
}
"""

LIST_PIPELINES_FOR_PROJECT_QUERY = """\
query pipelines($id: String!) {
  project(handleName: $id) {
    pipelines {
      path
      id
      description
      isSynced
    }
  }
}
"""

LIST_PIPELINES_FOR_PROJECT_FULL_QUERY = """\
query pipelines($id: String!) {
  project(handleName: $id) {
    pipelines {
      path
      id
      isSynced
      description
      seeders
      generators
    }
  }
}
"""

RUN_PIPELINE_MUTATION = """\
mutation run_pipeline($input: RunPipelineInput!){
  runPipeline(input: $input) {
    errors {
        code
        message
        fields
    }
    result {
        id
        jobStatus
    }
  }
}
"""

SWEEP_PIPELINE_MUTATION = """\
mutation sweep_pipeline($input: SweepPipelineInput!){
  sweepPipeline(input: $input) {
    errors {
        code
        message
        fields
    }
    result {
        id
        jobStatus
    }
  }
}
"""

GET_PIPELINE_BY_ID_QUERY = """\
query get_pipeline($id: ID!) {
    pipeline(id: $id) {
        path
        project { id }
    }
}
"""

GET_PIPELINE_BY_PATH_QUERY = """\
query get_pipeline($path: String!, $project_id: String!) {
    project(handleName: $project_id) {
        pipelines(limit: 1, filter: { path: { eq: $path } }) {
           id
        }
    }
}
"""

LIST_PIPELINE_RUNS_BY_PROJECT = """\
query get_pipeline_runs($id: String!) {
    project(handleName: $id) {
        pipelines {
            path
            pipelineRuns {
                id
                jobStatus
                dataset { path }
            }
        }
    }
}
"""

LIST_PIPELINE_RUNS_BY_PIPELINE = """\
query get_pipeline_runs($id: ID!) {
    pipeline(id: $id) {
        path
        pipelineRuns {
            id
            jobStatus
            dataset { path }
        }
    }
}
"""

GET_PIPELINE_RUN = """\
query get_pipeline_run($id: ID!) {
    pipelineRun(id: $id) {
        dataset { id }
        jobStatus
        provenance
        parameters
    }
}
"""

GET_PIPELINE_SWEEP = """\
query get_pipeline_sweep($id: ID!) {
    pipelineSweep(id: $id) {
        jobStatus
        pipelineRuns {
            id
            dataset { id }
            jobStatus
            provenance
            parameters
        }
    }
}
"""


class ApeClient(Base):
    """
    Client providing access to the APE service.
    """
    def list_transforms(self, project_uuid=None):
        """
        Returns a list of available transforms.
        """

        query_response = self.raw_query(query=LIST_TRANSFORMS_QUERY, variables={
            'id': project_uuid
        })

        return query_response['data']['project']['transforms']

    def get_transform(self, uuid=None, path=None, project_uuid=None):
        """
        Returns more detail about the transform.
        """
        if uuid is not None:
            query_response = self.raw_query(query=GET_TRANSFORM_BY_ID_QUERY, variables={
                'id': uuid
            })

            return query_response['data']['transform']
        elif path is not None and path.find(':') == -1 and project_uuid is not None:
            query_response = self.raw_query(query=GET_TRANSFORM_BY_PATH_QUERY, variables={
                'path': path,
                'project': project_uuid
            })
            if query_response['data']['project']['transforms']:
                return query_response['data']['project']['transforms'][0]
        elif path is not None and path.find(':') != -1:
            bits = path.split(':', maxsplit=1)
            query_response = self.raw_query(query=GET_TRANSFORM_BY_PATH_QUERY, variables={
                'path': bits[1],
                'project': bits[0]
            })
            if query_response['data']['project']['transforms']:
                return query_response['data']['project']['transforms'][0]
        return None

    def create_transform(self,
                         project_uuid=None,
                         name=None,
                         path=None,
                         inputs=None,
                         outputs=None,
                         implementation=None,
                         description=None,
                         configuration_schema=None,
                         type=None):
        """
        Create a transform. This will do some light checking before sending the
        request. More extensive checking is done on the server.
        """
        if project_uuid is None:
            raise AttributeError("transforms must be assigned to a project.")
        if type not in ["seeder", "mapper", "reducer"]:
            raise AttributeError("transform type must be one of seeder, mapper, or reducer.")
        if implementation is None:
            raise AttributeError("transform implementation must be provided.")
        if path is None:
            raise AttributeError("transform path must be provided.")
        if configuration_schema is not None:
            jsonschema.Draft4Validator.check_schema(configuration_schema)
        variables = {
            'input': {
                'project': project_uuid,
                'type': type,
                'implementation': implementation,
                'path': path,
                'name': name,
                'configurationSchema': dumps(configuration_schema)
            }
        }
        if outputs is not None:
            variables['input']['outputs'] = dumps(outputs)
        if inputs is not None:
            variables['input']['inputs'] = dumps(inputs)
        if description is not None:
            variables['input']['description'] = description
        result = self.raw_query(query=CREATE_TRANSFORM_MUTATION, variables=variables)

        return result['data']['createTransform']['result']

    def sync_transform(self,
                       project_uuid=None,
                       name=None,
                       path=None,
                       inputs=None,
                       outputs=None,
                       implementation=None,
                       description=None,
                       configuration_schema=None,
                       type=None):
        """
        Sync a transform. This will do some light checking before sending the
        request. More extensive checking is done on the server.
        """
        if project_uuid is None:
            raise AttributeError("transforms must be assigned to a project.")
        if type not in ["seeder", "mapper", "reducer"]:
            raise AttributeError("transform type must be one of seeder, mapper, or reducer.")
        if implementation is None:
            raise AttributeError("transform implementation must be provided.")
        if path is None:
            raise AttributeError("transform path must be provided.")
        if configuration_schema is not None:
            jsonschema.Draft4Validator.check_schema(configuration_schema)
        variables = {
            'input': {
                'project': project_uuid,
                'type': type,
                'implementation': implementation,
                'path': path,
                'name': name,
                'configurationSchema': dumps(configuration_schema)
            }
        }
        if outputs is not None:
            variables['input']['outputs'] = dumps(outputs)
        if inputs is not None:
            variables['input']['inputs'] = dumps(inputs)
        if description is not None:
            variables['input']['description'] = description
        result = self.raw_query(query=SYNC_TRANSFORM_MUTATION, variables=variables)

        return result['data']['syncTransform']['result']

    def update_transform(self,
                         project_uuid=None,
                         name=None,
                         path=None,
                         inputs=None,
                         outputs=None,
                         implementation=None,
                         description=None,
                         configuration_schema=None,
                         type=None):
        """
        Update a transform. This will do some light checking before sending the
        request. More extensive checking is done on the server.
        """
        if project_uuid is None:
            raise AttributeError("transforms must be assigned to a project.")
        if type not in ["seeder", "mapper", "reducer"]:
            raise AttributeError("transform type must be one of seeder, mapper, or reducer.")
        if implementation is None:
            raise AttributeError("transform implementation must be provided.")
        if path is None:
            raise AttributeError("transform path must be provided.")
        if configuration_schema is not None:
            jsonschema.Draft4Validator.check_schema(configuration_schema)
        variables = {
            'input': {
                'type': type,
                'implementation': implementation,
                'path': path,
                'name': name,
                'configurationSchema': dumps(configuration_schema)
            }
        }
        if outputs is not None:
            variables['input']['outputs'] = dumps(outputs)
        if inputs is not None:
            variables['input']['inputs'] = dumps(inputs)
        if description is not None:
            variables['input']['description'] = description
        algo = self.get_transform(path=path, project_uuid=project_uuid)
        if algo is None:
            return {}
        variables['id'] = algo['id']
        result = self.raw_query(query=UPDATE_TRANSFORM_MUTATION, variables=variables)

        return result['data']['updateTransform']

    def delete_transform(self, id, project_uuid=None):
        """
        Deletes the transform.
        """
        if self.is_uuid(id):
            variables = {
                'id': id
            }
        else:
            algo = self.get_transform(path=id, project_uuid=project_uuid)
            if algo is None:
                return {}
            variables = {'id': algo['id']}

        result = self.raw_query(query=DELETE_TRANSFORM_MUTATION, variables=variables)

        return result['data']['deleteTransform']

    def delete_synced_transform(self, id, project_uuid=None):
        """
        Deletes the transform.
        """
        if self.is_uuid(id):
            variables = {
                'id': id
            }
        else:
            algo = self.get_transform(path=id, project_uuid=project_uuid)
            if algo is None:
                return {}
            variables = {'id': algo['id']}

        result = self.raw_query(query=DELETE_SYNCED_TRANSFORM_MUTATION, variables=variables)

        return result['data']['deleteSyncedTransform']

    def list_generators(self, project_uuid=None):
        """
        List generators given a transform or project scope. If no scope, then all
        generators for an account.
        """
        if project_uuid is None:
            raise AttributeError("Project id must be provided.")
        query = LIST_GENERATORS_FOR_PROJECT_QUERY
        variables = {'id': project_uuid}

        query_response = self.raw_query(query=query, variables=variables)

        if 'errors' in query_response:
            return []
        return query_response['data']['project']['generators']

    def create_generator(self,
                         project_uuid=None,
                         path=None,
                         description=None,
                         transform=None,
                         transform_configuration=None,
                         extract=None,
                         load=None):
        """
        Creates a generator. This will do some checking before sending the request to make sure
        the request is reasonable.
        """
        transform_rec = self.get_transform(project_uuid=project_uuid, path=transform)
        if transform_rec is None:
            raise ValueError(f"The transform ({transform}) is not defined.")
        if transform_rec['type'] != 'seeder' and extract is None:
            raise ValueError("Non-seeding generators require an 'extract' query.")
        jsonschema.validate(instance=transform_configuration, schema=transform_rec['configurationSchema'])

        variables = {
            'input': {
                'path': path,
                'project': project_uuid,
                'transform': transform,
                'configuration': dumps(transform_configuration),
                'extract': extract,
                'load': load
            }
        }
        if description is not None:
            variables['input']['generator']['description'] = description
        result = self.raw_query(query=CREATE_GENERATOR_MUTATION, variables=variables)

        return result['data']['createGenerator']

    def sync_generator(self,
                       project_uuid=None,
                       path=None,
                       description=None,
                       transform=None,
                       transform_configuration=None,
                       extract=None,
                       load=None):
        """
        Syncs a generator. This will do some checking before sending the request to make sure
        the request is reasonable.
        """
        transform_rec = self.get_transform(project_uuid=project_uuid, path=transform)
        if transform_rec is None:
            raise ValueError(f"The transform ({transform}) is not defined.")
        if transform_rec['type'] != 'seeder' and extract is None:
            raise ValueError("Non-seeding generators require an 'extract' query.")
        jsonschema.validate(instance=transform_configuration, schema=transform_rec['configurationSchema'])

        variables = {
            'input': {
                'path': path,
                'project': project_uuid,
                'transform': transform,
                'configuration': dumps(transform_configuration),
                'extract': extract,
                'load': load
            }
        }
        if description is not None:
            variables['input']['generator']['description'] = description

        result = self.raw_query(query=SYNC_GENERATOR_MUTATION, variables=variables)

        return result['data']['syncGenerator']

    def update_generator(self,
                         project_uuid=None,
                         path=None,
                         description=None,
                         transform=None,
                         transform_configuration=None,
                         extract=None,
                         load=None):
        """
        Creates a generator. This will do some checking before sending the request to make sure
        the request is reasonable.
        """
        transform_rec = self.get_transform(project_uuid=project_uuid, path=transform)
        if transform_rec is None:
            raise ValueError(f"The transform ({transform}) is not defined.")
        if transform_rec['type'] != 'seeder' and extract is None:
            raise ValueError("Non-seeding generators require an 'extract' query.")
        jsonschema.validate(instance=transform_configuration, schema=transform_rec['configurationSchema'])

        variables = {
            'input': {
                'path': path,
                'transform': transform,
                'configuration': dumps(transform_configuration),
                'extract': extract,
                'load': load
            }
        }
        if description is not None:
            variables['input']['generator']['description'] = description
        generator = self.get_generator(path=path, project_uuid=project_uuid)
        if generator is None:
            return {}
        variables['id'] = generator['id']
        result = self.raw_query(query=UPDATE_GENERATOR_MUTATION, variables=variables)

        return result['data']['updateGenerator']

    def get_generator(self, uuid=None, path=None, project_uuid=None):
        """
        Returns more detail about the generator.
        """
        if uuid is not None:
            query_response = self.raw_query(query=GET_GENERATOR_BY_ID_QUERY, variables={
                'id': uuid
            })

            return query_response['data']['generator']
        elif path is not None and project_uuid is not None:
            query_response = self.raw_query(query=GET_GENERATOR_BY_PATH_QUERY, variables={
                'path': path,
                'project_id': project_uuid
            })
            return query_response['data']['project']['generators'][0]
        return None

    def delete_generator(self, id, project_uuid=None):
        """
        Deletes the generator.
        """
        if self.is_uuid(id):
            variables = {
                'id': id
            }
        else:
            generator = self.get_generator(path=id, project_uuid=project_uuid)
            if generator is None:
                return {}
            variables = {'id': generator['id']}

        result = self.raw_query(query=DELETE_GENERATOR_MUTATION, variables=variables)

        return result['data']['deleteGenerator']

    def delete_synced_generator(self, id, project_uuid=None):
        """
        Deletes a synced generator.
        """
        if self.is_uuid(id):
            variables = {
                'id': id
            }
        else:
            generator = self.get_generator(path=id, project_uuid=project_uuid)
            if generator is None:
                return {}
            variables = {'id': generator['id']}

        result = self.raw_query(query=DELETE_SYNCED_GENERATOR_MUTATION, variables=variables)

        return result['data']['deleteSyncedGenerator']

    def list_pipelines(self, project_uuid=None, all=False):
        """
        List pipelines given a project scope. If no scope, then all
        pipelines for an account.
        """

        variables = {}
        query = LIST_PIPELINES_FOR_PROJECT_QUERY
        if all:
            query = LIST_PIPELINES_FOR_PROJECT_FULL_QUERY
        variables['id'] = project_uuid

        query_response = self.raw_query(query=query, variables=variables)

        if 'errors' in query_response:
            return []

        return query_response['data']['project']['pipelines']

    def create_pipeline(self, project_uuid=None, description=None, path=None, seeders=None, generators=None):
        """
        Creates a pipeline.
        """
        variables = {
            'input': {
                'path': path,
                'project': project_uuid,
                'seeders': [],
                'generators': []
            }
        }

        if generators is not None:
            variables['input']['generators'] = [
                self._marshal_generator(label, generator) for label, generator in generators.items()]
        if seeders is not None:
            variables['input']['seeders'] = [self._marshal_seeder(seeder) for seeder in seeders]
        if description is not None:
            variables['input']['description'] = description

        result = self.raw_query(query=CREATE_PIPELINE_MUTATION, variables=variables)
        if 'errors' in result:
            return {"accepted": False}

        return result['data']['createPipeline']

    def sync_pipeline(self, project_uuid=None, description=None, path=None, seeders=None, generators=None):
        """
        Syncs a pipeline.
        """
        variables = {
            'input': {
                'path': path,
                'project': project_uuid,
                'seeders': [],
                'generators': []
            }
        }

        if generators is not None:
            variables['input']['generators'] = [
                self._marshal_generator(label, generator) for label, generator in generators.items()]
        if seeders is not None:
            variables['input']['seeders'] = [self._marshal_seeder(seeder) for seeder in seeders]
        if description is not None:
            variables['input']['description'] = description

        result = self.raw_query(query=SYNC_PIPELINE_MUTATION, variables=variables)
        if 'errors' in result:
            return {"accepted": False}

        return result['data']['syncPipeline']

    def update_pipeline(self, project_uuid=None, description=None, path=None, seeders=None, generators=None):
        """
        Updates a pipeline.
        """
        variables = {
            'input': {
                'path': path,
                'seeders': [],
                'generators': []
            }
        }

        if generators is not None:
            variables['input']['generators'] = [
                self._marshal_generator(label, generator) for label, generator in generators.items()]
        if seeders is not None:
            variables['input']['seeders'] = [self._marshal_seeder(seeder) for seeder in seeders]
        if description is not None:
            variables['input']['description'] = description
        pipeline = self.get_pipeline(path=path, project_uuid=project_uuid)
        if pipeline is None:
            return {}
        variables['id'] = pipeline['id']
        result = self.raw_query(query=UPDATE_PIPELINE_MUTATION, variables=variables)
        if 'errors' in result:
            return {"accepted": False}

        return result['data']['updatePipeline']

    def get_pipeline(self, uuid=None, path=None, project_uuid=None):
        """
        Returns more detail about the pipeline.
        """
        if uuid is not None:
            query_response = self.raw_query(query=GET_PIPELINE_BY_ID_QUERY, variables={
                'id': uuid
            })

            return query_response['data']['pipeline']
        elif path is not None and project_uuid is not None:
            query_response = self.raw_query(query=GET_PIPELINE_BY_PATH_QUERY, variables={
                'path': path,
                'project_id': project_uuid
            })
            return query_response['data']['project']['pipelines'][0]
        return None

    def get_pipeline_run(self, uuid):
        """
        Returns more detail about the pipeline run.
        """
        query_response = self.raw_query(query=GET_PIPELINE_RUN, variables={
            'id': uuid
        })
        return query_response['data']['pipelineRun']

    def get_pipeline_sweep(self, uuid):
        """
        Returns more detail about the pipeline sweep.
        """
        query_response = self.raw_query(query=GET_PIPELINE_SWEEP, variables={
            'id': uuid
        })
        return query_response['data']['pipelineSweep']

    def list_pipeline_runs(self, pipeline_id=None, project_uuid=None):
        """
        List pipeline runs for a pipeline.
        """
        if pipeline_id is not None:
            if self.is_uuid(pipeline_id):
                variables = {'id': pipeline_id}
            else:
                pipeline = self.get_pipeline(path=pipeline_id, project_uuid=project_uuid)
                if pipeline is None:
                    return []
                variables = {'id': pipeline['id']}
            query_response = self.raw_query(query=LIST_PIPELINE_RUNS_BY_PIPELINE, variables=variables)
            pipeline_path = query_response['data']['pipeline']['path']
            return [{'pipeline': pipeline_path, **run} for run in query_response['data']['pipeline']['pipelineRuns']]
        else:
            variables = {'id': project_uuid}
            query_response = self.raw_query(query=LIST_PIPELINE_RUNS_BY_PROJECT, variables=variables)
            return [
                {
                    'pipeline': pipeline['path'],
                    'dataset': self._get_dataset_path(run['dataset']),
                    'jobStatus': run['jobStatus']
                }
                for pipeline in query_response['data']['project']['pipelines']
                for run in pipeline['pipelineRuns']
            ]

    @staticmethod
    def _get_dataset_path(dataset):
        try:
            path = dataset['path']
        except (KeyError, TypeError):
            path = ''
        return path

    def _marshal_generator(self, label, values):
        new_values = self._marshal_seeder(values)
        new_values['label'] = label
        if 'dependsOn' not in new_values:
            new_values['dependsOn'] = []
        return new_values

    def _marshal_seeder(self, values):
        new_values = values.copy()
        if 'generator' in values and 'configuration' in values['generator']:
            new_values['generator'] = values['generator'].copy()
            new_values['generator']['configuration'] = dumps(values['generator']['configuration'])
        return new_values

    def delete_pipeline(self, id, project_uuid=None):
        """
        Deletes the pipeline.
        """
        if self.is_uuid(id):
            variables = {
                'id': id
            }
        else:
            pipeline = self.get_pipeline(path=id, project_uuid=project_uuid)
            if pipeline is None:
                return {}
            variables = {'id': pipeline['id']}

        result = self.raw_query(query=DELETE_PIPELINE_MUTATION, variables=variables)

        return result['data']['deletePipeline']

    def delete_synced_pipeline(self, id, project_uuid=None):
        """
        Deletes the synced pipeline.
        """
        if self.is_uuid(id):
            variables = {
                'id': id
            }
        else:
            pipeline = self.get_pipeline(path=id, project_uuid=project_uuid)
            if pipeline is None:
                return {}
            variables = {'id': pipeline['id']}

        result = self.raw_query(query=DELETE_SYNCED_PIPELINE_MUTATION, variables=variables)

        return result['data']['deleteSyncedPipeline']

    def run_pipeline(self, id, parameters=None, sweep_parameters=None, project_uuid=None):
        """
        Runs the pipeline.
        """
        if self.is_uuid(id):
            variables = {
                'input': {
                    'pipelineId': id
                }
            }
        else:
            pipeline = self.get_pipeline(path=id, project_uuid=project_uuid)
            if pipeline is None:
                return {}
            variables = {
                'input': {
                    'pipelineId': pipeline['id']
                }
            }
        query = RUN_PIPELINE_MUTATION
        result_key = 'runPipeline'
        if parameters is not None:
            variables['input']['parameters'] = dumps(parameters)
        if sweep_parameters is not None and any(sweep_parameters):
            variables['input']['sweptParameters'] = dumps(sweep_parameters)
            query = SWEEP_PIPELINE_MUTATION
            result_key = 'sweepPipeline'
        result = self.raw_query(query=query, variables=variables)
        return result['data'][result_key]



from cement import Controller, ex
import json
import sys
import yaml

from ..core.services import client


class Send(Controller):
    """
    Provides the 'send' verb.
    """

    class Meta:
        label = 'send'
        stacked_on = 'base'
        stacked_type = 'nested'
        help = 'send something'

    @ex(help='send a message to an agent',
        arguments=[
            (['id'], {'help': 'agent instance identifier', 'action': 'store'}),
            (['events'], {'help': 'file containing message(s) to send', 'action': 'store'})
        ])
    def agent_event(self):
        planning_client = client('planning', self.app)
        file = self.app.pargs.events
        if file == '-':
            # we assume that this is a one-shot set of events, so we read until STDIN closes
            content = sys.stdin.read()
            # read with YAML since that's a superset of JSON
            events = self._yaml_load(content)
        else:
            if file.endswith('.yaml') or file.endswith('.yml'):
                loader = self._yaml_load
            elif file.endswith('.json'):
                loader = json.load
            else:
                print("Only YAML and JSON files may be uploaded.")
                sys.exit(1)
            with open(file, 'r', encoding='utf-8') as fd:
                events = loader(fd)

        if not isinstance(events, list):
            events = [events]
        for event in events:
            result = planning_client.send_event_to_agent(self.app.pargs.id, event)
            self._print(result)

    @staticmethod
    def _yaml_load(fd):
        return list(yaml.load_all(fd, Loader=yaml.Loader))

    def _print(self, data):
        if self.app.pargs.output == 'json':
            print(json.dumps(data))
        else:
            print(yaml.dump(data))

    @property
    def _project_scope(self):
        return self.app.pargs.project or self.app.config.get('ursactl', 'project')



from cement import Controller, ex
import json
import yaml

from ..core.services import client


class Delete(Controller):
    """
    Provides the 'delete' verb.
    """

    class Meta:
        label = 'delete'
        stacked_on = 'base'
        stacked_type = 'nested'
        help = 'delete something'

    @ex(help='delete an agent',
        arguments=[
            (['--project'], {'help': 'project identifier', 'dest': 'project', 'action': 'store'}),
            (['agent'], {'help': 'agent uuid or path', 'action': 'store'})
        ])
    def agent(self):
        planning_client = client('planning', self.app)
        result = planning_client.delete_agent(self.app.pargs.agent, project_uuid=self._project_scope)
        self._print(result)

    @ex(help='delete a transform',
        arguments=[
            (['--project'], {'help': 'project identifier', 'dest': 'project', 'action': 'store'}),
            (['transform'], {'help': 'transform uuid or path', 'action': 'store'})
        ])
    def transform(self):
        ape_client = client('ape', self.app)

        result = ape_client.delete_transform(self.app.pargs.transform, project_uuid=self._project_scope)
        self._print(result)

    @ex(help='delete a dataset',
        arguments=[
            (['--project'], {'help': 'project identifier', 'dest': 'project', 'action': 'store'}),
            (['dataset_uuid'], {'help': 'The UUID of the dataset to delete'})
        ])
    def dataset(self):
        dss_client = client('dss', self.app)
        result = dss_client.delete_dataset(self.app.pargs.dataset_uuid, project_uuid=self._project_scope)
        self._print(result)

    @ex(help='delete a generator',
        arguments=[
            (['--project'], {'help': 'project identifier', 'dest': 'project', 'action': 'store'}),
            (['generator'], {'help': 'generator uuid or path', 'action': 'store'})
        ])
    def generator(self):
        ape_client = client('ape', self.app)

        result = ape_client.delete_generator(self.app.pargs.generator, project_uuid=self._project_scope)
        self._print(result)

    @ex(help='delete a package',
        arguments=[
            (['--project'], {'help': 'project identifier', 'dest': 'project', 'action': 'store'}),
            (['package'], {'help': 'package uuid or path', 'action': 'store'})
        ])
    def package(self):
        planning_client = client('planning', self.app)
        result = planning_client.delete_package(self.app.pargs.package, project_uuid=self._project_scope)
        self._print(result)

    @ex(help='delete pipeline',
        arguments=[
            (['--project'], {'help': 'project identifier', 'dest': 'project', 'action': 'store'}),
            (['pipeline'], {'help': 'pipeline uuid or path', 'action': 'store'})
        ])
    def pipeline(self):
        ape_client = client('ape', self.app)
        pipeline = self.app.pargs.pipeline
        result = ape_client.delete_pipeline(pipeline, project_uuid=self._project_scope)
        self._print(result)

    def _print(self, data):
        if self.app.pargs.output == 'json':
            print(json.dumps(data))
        else:
            print(yaml.dump(data))

    @property
    def _project_scope(self):
        return self.app.pargs.project or self.app.config.get('ursactl', 'project')



"""
Utilities for working with transforms in the CLI.
"""
import json
import yaml
import sys


def load_transform(file):
    if file.endswith('.yaml') or file.endswith('.yml'):
        loader = _yaml_load
    elif file.endswith('.json'):
        loader = json.load
    elif file.endswith('.md'):
        loader = load_from_markdown
    elif file.endswith('.lua'):
        loader = load_from_lua
    else:
        print("Only YAML, JSON, Lua, and Markdown files may be uploaded.")
        sys.exit(1)

    with open(file, 'r', encoding='utf-8') as fd:
        documents = loader(fd)
    return documents


def _yaml_load(fd):
    return list(yaml.load_all(fd, Loader=yaml.Loader))


def load_from_lua(source):  # noqa: C901
    import lupa.lua53

    info = {'description': '', 'configurationSchema': None}
    content = list(source)

    lua = lupa.lua53.LuaRuntime()
    try:
        lua.execute("".join(content))
    except lupa.lua53.LuaSyntaxError as e:
        print(e)
        return

    # find top comment in file
    first = None
    last = None
    for i in range(0, len(content)-1):
        if first is None and content[i].startswith('--[['):
            first = i
        elif first is not None and last is None and (
                content[i].startswith(']]') or content[i].startswith('--]]')):
            last = i
            break
    if first is None or last is None:
        print("No top-level comment found in file.")
        return

    lines = content[first+1:last-1]
    name, lines = _get_name(lines)
    info['name'] = name

    for line in lines:
        line = line.strip()
        if line.startswith('@'):
            parts = line[1:].split(' ', maxsplit=1)
            if len(parts) == 2:
                if parts[0] in ['input', 'output']:
                    bits = parts[1].split(' ', maxsplit=1)
                    try:
                        info[parts[0] + 's'][bits[0]] = bits[1]
                    except KeyError:
                        info[parts[0] + 's'] = {bits[0]: bits[1]}
                else:
                    info[parts[0]] = parts[1]
        else:
            info['description'] += line + '\n'

    if lua.globals().configurationSchema is not None:
        info['configurationSchema'] = _table_to_dict(lua.globals().configurationSchema)

    info['implementation'] = ''.join(content)
    return info


def _table_to_dict(lua_value):
    if lua_value is None:
        return None
    if isinstance(lua_value, (str, bool, int, float)):
        return lua_value
    d = {k: _table_to_dict(v) for k, v in lua_value.items()}
    if all([isinstance(n, int) for n in d.keys()]):
        return [v for _, v in sorted(d.items())]
    else:
        return d


def load_from_markdown(source):
    info = {}
    content = list(source)
    if content[0] == "---\n":
        split_at = len(content)
        # get metadata from header
        for i in range(1, len(content)-1):
            if content[i] == "---\n":
                split_at = i
                break
        raw_yaml = "".join(content[1:split_at])
        content = content[split_at+1:]
        info = yaml.load(raw_yaml, Loader=yaml.Loader)

    info['configurationSchema'] = None
    info['description'] = ''

    name, content = _get_name(content)
    chunks = _split_into_chunks(content)

    # first chunk is the description if it doesn't have a '##' at the start
    if not chunks[0][0].startswith('## '):
        info['description'] = "".join(chunks[0])
        chunks = chunks[1:]

    info['name'] = name

    for chunk in chunks:
        _update_info_from_chunk(info, chunk)

    return info


def _update_info_from_chunk(info, chunk):
    if chunk[0].startswith('## Metadata'):
        info.update(yaml.load("".join(chunk[1:]), Loader=yaml.Loader))
    if chunk[0].startswith('## Configuration Schema'):
        # we need the ```yaml ... ``` parts (or json)
        source = _find_fenced_block(chunk[1:], "yaml")
        if source is not None:
            info['configurationSchema'] = yaml.load("".join(source), Loader=yaml.Loader)
        else:
            source = _find_fenced_block(chunk[1:], "json")
            if source is not None:
                info['configurationSchema'] = json.loads("".join(source))
    if chunk[0].startswith('## Implementation'):
        # we need the ```lua ... ``` parts
        lua = _find_fenced_block(chunk[1:], "lua")
        if lua is not None:
            info['implementation'] = "".join(lua)


def _get_name(content):
    name = None
    # the first top-level header is the name of the transform
    for i in range(0, len(content)-1):
        if content[i].startswith('# '):
            name = content[i][2:].strip()
            content = content[i+1:]
            break
    return name, content


def _split_into_chunks(content):
    chunks = []
    chunk = []
    for line in content:
        if line.startswith('## '):
            if len(chunk) > 0:
                chunks.append(chunk)
            chunk = []
        chunk.append(line)
    if len(chunk) > 0:
        chunks.append(chunk)
    return chunks


def _find_fenced_block(chunk, tag):
    buffer = []
    started = False
    for line in chunk:
        if not started and line == f'```{tag}\n':
            started = True
            continue
        if started and line == '```\n':
            return buffer
        if started:
            buffer.append(line)
    return None



from cement import Controller, ex
import json
import yaml

from ..core.services import client


class Stop(Controller):
    """
    Provides the 'stop' verb.
    """

    class Meta:
        label = 'stop'
        stacked_on = 'base'
        stacked_type = 'nested'
        help = 'stop something'

    @ex(help='stop an agent',
        arguments=[
            (['id'], {'help': 'agent instance identifier', 'action': 'store'})
        ])
    def agent(self):
        planning_client = client('planning', self.app)
        result = planning_client.stop_agent(self.app.pargs.id)
        self._print(result)

    def _print(self, data):
        if self.app.pargs.output == 'json':
            print(json.dumps(data))
        else:
            print(yaml.dump(data))



from ursactl.core.services import client
from ursactl.core._base import Base


class PipelineRun(Base):
    """
    Provides access to a pipeline run.
    """
    def __init__(self, *args, pipeline=None, **kwargs):
        super().__init__(*args, **kwargs)
        self.pipeline = pipeline

    @property
    def client(self):
        if self._client is None:
            self._client = client('ape', self.app)
        return self._client

    @property
    def job_status(self):
        return self._data['jobStatus']

    @property
    def provenance(self):
        return self._data['provenance']

    @property
    def parameters(self):
        return self._data['parameters']

    @property
    def dataset(self):
        if self._data['dataset']:
            return self.project.Dataset(uuid=self._data['dataset']['id'])

    def refresh(self):
        if self.uuid is not None:
            self._cached_data = self.client.get_pipeline_run(uuid=self.uuid)

    @property
    def _data(self):
        if self._cached_data is None:
            if self.uuid is None:
                self._cached_data = {
                    'dataset': None,
                    'jobStatus': None,
                    'provenance': None,
                    'parameters': None,
                }
            else:
                self._cached_data = self.client.get_pipeline_run(uuid=self.uuid)
        return self._cached_data



from cement import Controller, ex

from ..core.services import client


class Refresh(Controller):
    """
    Provides the 'refresh' verb.
    """
    class Meta:
        label = 'refresh'
        stacked_on = 'base'
        stacked_type = 'nested'
        help = 'refresh something'

    @ex(help="refresh token")
    def token(self):
        iam_client = client('iam', self.app)
        iam_client.clear_token()
        iam_client.get_token()



import sys

from cement import Controller, ex
import json
import yaml

from ..core.services import client
from ..core.utils.magic import magic
from ursactl.controllers.utils.transforms import load_transform


def _yaml_load(fd):
    return list(yaml.load_all(fd, Loader=yaml.Loader))


class Create(Controller):
    """
    Provides the 'create' verb.
    """

    class Meta:
        label = 'create'
        stacked_on = 'base'
        stacked_type = 'nested'
        help = 'create something'

    @ex(help='create agent',
        arguments=[
            (['file'], {'help': 'agent definition file', 'action': 'store'}),
            (['--project'], {'help': 'project identifier', 'dest': 'project', 'action': 'store'}),
        ])
    def agent(self):
        project_uuid = self._project_scope
        planning_client = client('planning', self.app)

        file = self.app.pargs.file
        if file.endswith('.yaml') or file.endswith('.yml'):
            loader = _yaml_load
        elif file.endswith('.json'):
            loader = json.load
        else:
            print("Only YAML and JSON files may be uploaded.")
            sys.exit(1)

        with open(file, 'r', encoding='utf-8') as fd:
            documents = loader(fd)

        if not isinstance(documents, list):
            documents = [documents]

        for content in documents:
            args = {
                'name': content['name'],
                'packages': content['packages'],
                'behaviors': content['behaviors'],
                'project_uuid': project_uuid,
            }

            result = planning_client.create_agent(**args)
            self._print(result)

    @ex(help='create transform',
        arguments=[
            (['file'], {'help': 'transform definition file', 'action': 'store'}),
            (['--project'], {'help': 'project identifier', 'dest': 'project', 'action': 'store'}),
        ])
    def transform(self):
        project_uuid = self._project_scope
        ape_client = client('ape', self.app)

        file = self.app.pargs.file
        documents = load_transform(file)

        if documents is None:
            print(f"Unable to load transform from {file}")
            sys.exit(1)

        if not isinstance(documents, list):
            documents = [documents]

        for content in documents:
            args = {
                'path': content['path'],
                'name': content['name'],
                'type': content['type'],
                'implementation': content['implementation'],
                'configuration_schema': content['configurationSchema'],
                'project_uuid': project_uuid,
            }
            if 'outputs' in content:
                args['outputs'] = content['outputs']
            if 'inputs' in content:
                args['inputs'] = content['inputs']
            if 'description' in content:
                args['description'] = content['description']

            result = ape_client.create_transform(**args)
            self._print(result)

    @ex(help='create dataset',
        arguments=[
            (['source'], {'help': 'dataset to upload', 'action': 'store'}),
            (['path'], {'help': 'path of the dataset', 'action': 'store'}),
            (['--project'], {'help': 'project identifier', 'dest': 'project', 'action': 'store'}),
            (['--type'], {'help': 'file type if pushing to the datastore', 'action': 'store', 'dest': 'file_type'})
         ])
    def dataset(self):
        project_uuid = self._project_scope
        path = self.app.pargs.path
        source = self.app.pargs.source
        dss_client = client('dss', self.app)
        file_type = self.app.pargs.file_type or magic(source)
        result = dss_client.create_dataset(project_uuid=project_uuid, path=path, data=source, data_type=file_type)
        self._print(result)

    @ex(help='create a new package from a file',
        arguments=[
            (['file'], {'help': 'package definition file', 'action': 'store'}),
            (['--project'], {'help': 'project identifier', 'dest': 'project', 'action': 'store'}),
        ])
    def package(self):
        project_uuid = self._project_scope
        planning_client = client('planning', self.app)

        file = self.app.pargs.file
        with open(file) as fd:
            # read the content of the file
            content = fd.read()

        result = planning_client.create_package(project_uuid=project_uuid, content=content)
        self._print(result)

    @ex(help='create a new project',
        arguments=[
            (['name'], {'help': 'project name', 'action': 'store'}),
            (['description'], {
                'help': 'project description', 'action': 'store'})
        ]
        )
    def project(self):
        projects_client = client('projects', self.app)
        result = projects_client.create(
            name=self.app.pargs.name, description=self.app.pargs.description)
        self._print(result)

    @ex(help='create a generator from a file',
        arguments=[
            (['file'], {'help': 'generator definition file', 'action': 'store'}),
            (['--project'], {'help': 'project identifier', 'dest': 'project', 'action': 'store'}),
        ])
    def generator(self):
        ape_client = client('ape', self.app)

        file = self.app.pargs.file
        if file.endswith('.yaml') or file.endswith('.yml'):
            loader = _yaml_load
        elif file.endswith('.json'):
            loader = json.load
        else:
            print("Only YAML and JSON files may be uploaded.")
            sys.exit(1)

        with open(file, 'r', encoding='utf-8') as fd:
            documents = loader(fd)

        if not isinstance(documents, list):
            documents = [documents]

        for content in documents:
            args = {
                'path': content['path'],
                'transform': content['transform']['path'],
                'transform_configuration': content['transform']['configuration'],
                'load': content['load'],
                'project_uuid': self._project_scope
            }

            for key in ['extract', 'description']:
                if key in content and content[key] is not None:
                    args[key] = content[key]

            result = ape_client.create_generator(**args)
            self._print(result)

    @ex(help='create a pipeline from a file',
        arguments=[
            (['file'], {'help': 'pipeline definition file', 'action': 'store'}),
            (['--project'], {'help': 'project identifier', 'dest': 'project', 'action': 'store'}),
        ])
    def pipeline(self):
        ape_client = client('ape', self.app)

        file = self.app.pargs.file
        if file.endswith('.yaml') or file.endswith('.yml'):
            loader = _yaml_load
        elif file.endswith('.json'):
            loader = json.load
        else:
            print("Only YAML and JSON files may be uploaded.")
            sys.exit(1)

        with open(file, 'r', encoding='utf-8') as fd:
            documents = loader(fd)

        if not isinstance(documents, list):
            documents = [documents]

        for content in documents:
            args = {
                'path': content['path'],
                'project_uuid': self._project_scope
            }

            for key in ['seeders', 'generators', 'description']:
                if key in content and content[key] is not None:
                    args[key] = content[key]

            result = ape_client.create_pipeline(**args)
            self._print(result)

    def _print(self, data):
        if self.app.pargs.output == 'json':
            print(json.dumps(data))
        else:
            print(yaml.dump(data))

    @property
    def _project_scope(self):
        return self.app.pargs.project or self.app.config.get('ursactl', 'project')



from ursactl.core.services import client
from ursactl.core._base import Base
from .pipeline_run import PipelineRun
from .pipeline_sweep import PipelineSweep


class Pipeline(Base):
    """
    Provides access to a pipeline definition.
    """
    @property
    def client(self):
        if self._client is None:
            self._client = client('ape', self.app)
        return self._client

    # @property
    # def project(self):
    #     from ursactl.core.project import Project

    #     return Project(self._data['project_uuid'], app=self.app)

    @property
    def name(self):
        return self._data['name']

    @property
    def description(self):
        return self._data['description']

    @property
    def pipeline_runs(self):
        return []

    @property
    def pipeline_sweeps(self):
        return []

    def run(self, parameters=None):
        """
        Runs the pipeline, returning a PipelineRun object.
        """
        result = self.client.run_pipeline(
            self.uuid,
            parameters=parameters,
            project_uuid=self.project.uuid)

        return PipelineRun(pipeline=self, project=self.project, uuid=result['id'], app=self.app, client=self.client)

    def sweep(self, parameters=None, sweep_parameters=None):
        """
        Sweeps the pipeline, returning a PipelineSweep object.
        """
        result = self.client.run_pipeline(
            self.uuid,
            parameters=parameters,
            sweep_parameters=sweep_parameters,
            project_uuid=self.project.uuid)

        return PipelineSweep(
            pipeline=self,
            project=self.project,
            uuid=result['result']['id'],
            app=self.app,
            client=self.client)

    @property
    def _data(self):
        if self._cached_data is None:
            if self.uuid is None:
                self._cached_data = {
                    'path': None,
                    'description': None,
                    'seeders': [],
                    'generators': []
                }
            else:
                self._cached_data = self.client.get_pipeline(uuid=self.uuid)
        return self._cached_data



from pathlib import Path
import sys

from cement import Controller
import json
import yaml

from ..core.services import client
import ursactl.controllers.utils.transforms


class Sync(Controller):
    """
    Provides the 'sync' verb
    """

    class Meta:
        label = 'sync'
        stacked_on = 'base'
        stacked_type = 'nested'
        help = 'synchronize a directory with Ursa Frontier'
        arguments = [
            (['--project'], {
                 'help': 'override default project to use with this repository',
                 'dest': 'project',
                 'action': 'store'}),
            (['--dir', '--directory'], {
                 'help': 'directory to initialize (defaults to \'.\')',
                 'dest': 'directory',
                 'action': 'store',
                 'default': '.'}),
            (['--dry-run'], {
                    'help': 'rather than sync, list all resources that would be synced',
                    'dest': 'dry_run',
                    'action': 'store_true'})
        ]

    def _default(self):
        root = Path(self.app.pargs.directory)
        if not root.exists():
            print("Error: %s does not exist." % root)
            sys.exit(1)
        elif not root.is_dir():
            print("Error: %s exists and is not a directory." % root)
            sys.exit(1)
        ape_client = client('ape', self.app)

        resources = self._load_resources(root)

        if self.app.pargs.dry_run:
            print("Would sync:")
            print(yaml.dump({k: list(v.keys()) for k, v in resources.items() if len(v) > 0}))
            return

        # now we can sync resources
        self._sync_datasets(resources['dataset'])
        self._sync_transforms(ape_client, resources['transform'])
        self._sync_generators(ape_client, resources['generator'])
        self._sync_pipelines(ape_client, resources['pipeline'])

    def _sync_datasets(self, datasets):
        pass

    def _sync_transforms(self, client, transforms):
        for item in client.list_transforms(self._project_scope):
            if item['path'] not in transforms and item['isSynced']:
                client.delete_synced_transform(item['id'])
            elif item['path'] in transforms and not item['isSynced']:
                print(f"Unable to sync transform #{item['path']} because it is not synced.")
                transforms.delete(item['path'])
        for item in transforms.values():
            args = {
                'path': item['path'],
                'name': item['name'],
                'type': item['type'],
                'implementation': item['implementation'],
                'configuration_schema': item['configurationSchema'],
                'project_uuid': self._project_scope,
            }
            if 'outputs' in item:
                args['outputs'] = item['outputs']
            if 'inputs' in item:
                args['inputs'] = item['inputs']
            if 'description' in item:
                args['description'] = item['description']
            result = client.sync_transform(**args)
            self._print(result)

    def _sync_generators(self, client, generators):
        for item in client.list_generators(self._project_scope):
            if item['path'] not in generators and item['isSynced']:
                client.delete_synced_generator(item['id'])
            elif item['path'] in generators and not item['isSynced']:
                print(f"Unable to sync generator #{item['path']} because it is not synced.")
                generators.delete(item['path'])
        for item in generators.values():
            args = {
                'path': item['path'],
                'transform': item['transform']['path'],
                'transform_configuration': item['transform']['configuration'],
                'load': item['load'],
                'project_uuid': self._project_scope
            }

            for key in ['extract', 'description']:
                if key in item and item[key] is not None:
                    args[key] = item[key]

            result = client.sync_generator(**args)
            self._print(result)

    def _sync_pipelines(self, client, pipelines):
        for item in client.list_pipelines(self._project_scope):
            if item['path'] not in pipelines and item['isSynced']:
                client.delete_synced_pipeline(item['id'])
            elif item['path'] in pipelines and not item['isSynced']:
                print(f"Unable to sync pipeline #{item['path']} because it is not synced.")
                pipelines.delete(item['path'])
        for item in pipelines.values():
            args = {
                'path': item['path'],
                'project_uuid': self._project_scope
            }
            for key in ['seeders', 'generators', 'description']:
                if key in item and item[key] is not None:
                    args[key] = item[key]

            result = client.sync_pipeline(**args)
            self._print(result)

    def _load_resources(self, root):  # noqa: C901
        # we want to look at each resource we can find locally and compute a hash
        # that makes sense for the data stored on the server - to see if there's a difference
        # for now, we're focused on the data side of things
        # we'll need to look at the planning side of things later
        # we look in each of the directories created by `init`, but use them only to
        # establish the default resource type. Resources can have a `kind` metadata
        # attribute that we can use to determin the kind of resource we're looking at.
        # datasets are special, but the same file with the .yaml extension can house
        # the metadata (so foo.csv has a foo.csv.yaml metadata file)
        #
        # OR, for datasets, we can have an index.yaml file that lists all of the datasets
        # we want to sync along with their metadata. Or both, until we find out which one
        # users find easiest to use.
        resources = {
            'transform': {},
            'generator': {},
            'pipeline': {},
            'dataset': {},
        }

        for file in root.rglob("*"):
            if file.is_dir():
                continue
            if file.stat().st_size == 0:
                continue
            if file.name.startswith('.'):
                continue
            info = None
            suffix = file.suffix
            if suffix == '.md':
                # likely a transform
                with open(file, "r") as fd:
                    info = ursactl.controllers.utils.transforms.load_from_markdown(fd)
            elif suffix == '.lua':
                # likely a transform
                with open(file, "r") as fd:
                    info = ursactl.controllers.utils.transforms.load_from_lua(fd)
            elif suffix in ['.yaml', '.yml']:
                with open(file, "r") as fd:
                    info = list(yaml.load_all(fd, Loader=yaml.Loader))
            elif suffix == '.json':
                with open(file, "r") as fd:
                    info = json.load(fd)
            else:
                continue
            info = info if isinstance(info, list) else [info]
            for item in info:
                kind = self._get_kind(root, file, item)
                if kind in resources:
                    resources[kind][item['path']] = item

        return resources

    def _get_kind(self, root, file, info) -> str:
        if info is None:
            return 'unknown'
        if 'kind' in info and info['kind'] not in [None, '']:
            return info['kind']
        if file.suffix in ['.yaml', '.yml'] and file.with_name(file.stem).exists():
            return 'dataset'
        bits = file.relative_to(root).parts
        for type in ['transforms', 'generators', 'pipelines', 'datasets']:
            if type in bits:
                return type[:-1]
        return 'unknown'

    @property
    def _project_scope(self):
        return self.app.pargs.project or self.app.config.get('ursactl', 'project')

    def _print(self, data):
        print(yaml.dump(data))



from cement import Controller, ex
from json import dumps as json_dump
from yaml import dump as yaml_dump
from configparser import NoOptionError

from ursactl.core.services import client
from ursactl.core.exc import UrsaProjectNotDefined


class List(Controller):
    """
    Provides the `ursactl list ...` actions.
    """

    class Meta:
        label = 'list'
        stacked_on = 'base'
        stacked_type = 'nested'
        help = 'list something'

    @ex(help='list agents',
        arguments=[
            (['--project'], {'help': 'project identifier', 'dest': 'project', 'action': 'store'}),
        ])
    def agents(self):
        planning_client = client('planning', self.app)

        agents = planning_client.list_agents(project_uuid=self._project_scope)
        if self.app.pargs.output is None:
            data = []
            for item in agents:
                data.append(
                    [item['name'], item['id']]
                )
            data.sort()
            self.app.render(data, headers=['Name', 'UUID'])
        else:
            self._print(agents)

    @ex(help='list transforms',
        arguments=[
            (['--project'], {'help': 'project identifier', 'dest': 'project', 'action': 'store'}),
        ])
    def transforms(self):
        ape_client = client('ape', self.app)

        algos = ape_client.list_transforms(project_uuid=self._project_scope)

        if self.app.pargs.output is None:
            data = []
            for item in algos:
                data.append(
                    [item['path'], item['name'], item['type']]
                )
            data.sort()
            self.app.render(data, headers=['Path', 'Name', 'Type'])
        else:
            self._print(algos)

    @ex(help='list datasets',
        arguments=[
            (['--project'], {'help': 'project identifier', 'dest': 'project', 'action': 'store'}),
        ])
    def datasets(self):
        dss_client = client('dss', self.app)
        datasets = dss_client.list_datasets(self._project_scope)

        if self.app.pargs.output is None:
            data = []
            for item in datasets:
                data.append(
                    [item['path'], item['id'], item['size'], item['content_type']]
                )
            data.sort()
            self.app.render(data, headers=['Path', 'UUID', 'Size', 'Type'])
        else:
            self._print(list(datasets))

    @ex(help='list generators',
        arguments=[
            (['--project'], {'help': 'project identifier', 'dest': 'project', 'action': 'store'}),
        ])
    def generators(self):
        ape_client = client('ape', self.app)

        generators = ape_client.list_generators(project_uuid=self._project_scope)
        if self.app.pargs.output is None:
            data = []
            for item in generators:
                data.append(
                    [item['path'], item['id'], item['transform']['type'], item['transform']['path']]
                )
            data.sort()
            self.app.render(data, headers=['Path', 'UUID', 'Type', 'transform'])
        else:
            self._print(generators)

    @ex(help='list packages',
        arguments=[
            (['--project'], {'help': 'project identifier', 'dest': 'project', 'action': 'store'}),
        ])
    def packages(self):
        planning_client = client('planning', self.app)

        packages = planning_client.list_packages(project_uuid=self._project_scope)
        if self.app.pargs.output is None:
            data = []
            for item in packages:
                data.append(
                    [item['name'], item['id']]
                )
            data.sort()
            self.app.render(data, headers=['Name', 'UUID'])
        else:
            self._print(packages)

    @ex(help='list pipelines',
        arguments=[
            (['--project'], {'help': 'project identifier', 'dest': 'project', 'action': 'store'}),
        ])
    def pipelines(self):
        ape_client = client('ape', self.app)

        if self.app.pargs.output is None:
            pipelines = ape_client.list_pipelines(project_uuid=self._project_scope)
            data = []
            for item in pipelines:
                data.append(
                    [item['path'], item['id'], item['description']]
                )
            data.sort()
            self.app.render(data, headers=['Path', 'UUID', 'Description'])
        else:
            pipelines = ape_client.list_pipelines(
                project_uuid=self._project_scope,
                all=True)
            self._print(pipelines)

    @ex(help='list projects')
    def projects(self):
        projects_client = client('projects', self.app)
        projects = projects_client.list()

        if self.app.pargs.output is None:
            data = []
            for item in projects:
                data.append(
                    [item['name'], item['isArchived'], item['user']['handle'], item['description']])
            data.sort()
            self.app.render(data, headers=['Name', 'Archived?', 'Owner', 'Description'])
        else:
            self._print(list(projects))

    @ex(help='list running agents',
        arguments=[
            (['--project'], {'help': 'project identifier', 'dest': 'project', 'action': 'store'}),
            (['agent'], {'help': 'agent identifier', 'action': 'store', 'nargs': '?'})
        ])
    def running_agents(self):
        planning_client = client('planning', self.app)

        running_agents = planning_client.list_running_agents(agent_id=self.app.pargs.agent, project_uuid=self._project_scope)
        if self.app.pargs.output is None:
            data = []
            for item in running_agents:
                data.append(
                    [item['agent'], item['startedAt'], item['id'], item['status']]
                )
            data.sort()
            self.app.render(data, headers=['Agent', 'Started At', 'Instance UUID', 'Status'])
        else:
            self._print(running_agents)

    @ex(help='list pipeline runs',
        arguments=[
            (['--project'], {'help': 'project identifier', 'dest': 'project', 'action': 'store'}),
            (['pipeline'], {'help': 'pipeline identifier', 'action': 'store', 'nargs': '?'})
        ])
    def pipeline_runs(self):
        ape_client = client('ape', self.app)

        pipeline_runs = ape_client.list_pipeline_runs(pipeline_id=self.app.pargs.pipeline, project_uuid=self._project_scope)

        if self.app.pargs.output is None:
            data = []
            for item in pipeline_runs:
                data.append(
                    [item['pipeline'], item['jobStatus'], item['dataset']]
                )
            data.sort()
            self.app.render(data, headers=['Pipeline', 'Status', 'Dataset'])
        else:
            self._print(pipeline_runs)

    @property
    def _project_scope(self):
        try:
            scope = self.app.pargs.project or self.app.config.get('ursactl', 'project')
        except NoOptionError:
            raise UrsaProjectNotDefined("Project must be defined.")
        return scope

    def _print(self, data):
        if self.app.pargs.output == 'json':
            print(json_dump(data))
        else:
            print(yaml_dump(data))



from typing import Any

from airflow.exceptions import AirflowException
from airflow.hooks.base import BaseHook
from airflow.utils.log.logging_mixin import LoggingMixin

from ursactl.core.services import client


class ConfigManager:
    def __init__(self, config):
        self.sections = config

    def get(self, section, key):
        if section not in self.sections:
            return None
        if key not in self.sections[section]:
            return None
        return self.sections[section][key]


class CacheDb:
    def __init__(self):
        self.cache = []

    def insert(self, dict):
        self.cache.append(dict)

    def search(self, query):
        print(query)
        return self.cache

    def delete(self, query):
        print(query)
        self.cache = []


class UrsaCtlProviderApp:
    """
    This presents enough of the App interface for the services to work.
    """

    def __init__(self, api_key, api_secret, api_host, project):
        self.config = ConfigManager({
            "ursactl": {
                "api_key": api_key,
                "api_secret": api_secret,
                "platform": api_host,
                "project": project
            }
        })
        self.db = None

    def log(self, msg):
        self.log.info(msg)


class UrsaCtlHook(BaseHook, LoggingMixin):
    """
    Uses Ursa Frontier Platform API to send and receive datasets and run pipelines.
    """

    conn_name_attr = "ursa_frontier_conn_id"
    default_conn_name = "ursa_frontier_default"
    conn_type = "ursa_frontier_platform"
    hook_name = "Ursa Frontier"

    def __init__(self, ursa_frontier_conn_id: str = "ursa_frontier_default"):
        super().__init__()
        conn = self.get_connection(ursa_frontier_conn_id)
        self.api_key = conn.extra_dejson.get("api_key", None)
        self.api_secret = conn.extra_dejson.get("api_secret", None)
        self.api_host = conn.extra_dejson.get("platform", "app.ursafrontier.com")
        self.api_projoect = conn.extra_dejson.get("project", None)

        self.host = conn.host

        if self.api_key is None:
            raise AirflowException("api_key must be specified in the Ursa Frontier Platform connection details")

        if self.api_secret is None:
            raise AirflowException("api_secret must be specified in the Ursa Frontier Platform connection details")

        if self.api_project is None:
            raise AirflowException("project must be specified in the Ursa Frontier Platform connection details")

        self.log.info("Setting up api keys for Ursa Frontier Platform")
        app = UrsaCtlProviderApp(self.api_key, self.api_secret, self.api_host, self.api_project)
        self.dss_client = client('dss', app)
        self.ape_client = client('ape', app)

    def get_dataset(self, dataset_id, version=None):
        """
        Retrieves a dataset from the Ursa Frontier Platform.
        """
        self.log.info("Retrieving dataset %s", dataset_id)
        return self.dss_client.get_dataset(dataset_id, version)

    @staticmethod
    def get_connection_form_widgets() -> dict[str, Any]:
        """Returns connection widgets to add to connection form"""
        from flask_appbuilder.fieldwidgets import BS3PasswordFieldWidget, BS3TextFieldWidget
        from flask_babel import lazy_gettext
        from wtforms import PasswordField, StringField

        return {
            "api_host": StringField(lazy_gettext("API endpoint"), widget=BS3TextFieldWidget()),
            "api_key": StringField(lazy_gettext("API key"), widget=BS3TextFieldWidget()),
            "api_secret": PasswordField(lazy_gettext("API secret"), widget=BS3PasswordFieldWidget()),
            "api_project": StringField(lazy_gettext("Project"), widget=BS3TextFieldWidget()),
        }



import os

from cement import App, TestApp, init_defaults
from cement.core.exc import CaughtSignal
from cement.utils import fs
from pathlib import Path
from tinydb import TinyDB
from requests.exceptions import ConnectionError

from .controllers.base import Base
from .controllers.create import Create
from .controllers.delete import Delete
from .controllers.get import Get
from .controllers.init import Init
from .controllers.list import List
from .controllers.refresh import Refresh
from .controllers.run import Run
from .controllers.send import Send
from .controllers.show import Show
from .controllers.stop import Stop
from .controllers.sync import Sync
from .controllers.update import Update
from .core.exc import UrsaCtlError

# configuration defaults
CONFIG = init_defaults('ursactl')
CONFIG['ursactl']['cache_file'] = '~/.ursactl/.cache'
CONFIG['ursactl']['platform'] = 'https://app.ursafrontier.cloud/'


def extend_cache(app):
    app.log.info('extending with cache')
    db_file = app.config.get('ursactl', 'cache_file')

    # ensure that we expand the full path
    db_file = fs.abspath(db_file)
    app.log.info('cache file is: %s' % db_file)

    # ensure our parent directory exists
    db_dir = os.path.dirname(db_file)
    if not os.path.exists(db_dir):
        os.makedirs(db_dir)

    app.extend('db', TinyDB(db_file))


def exploded_paths():
    """
    Returns a list of directories based on the current working directory.
    Doesn't go past the home directory, if under the home directory.
    """
    cwd = os.getcwd()
    parts = cwd.split(os.path.sep)
    paths = []
    homedir = Path.home()

    for i in range(len(parts)):
        paths.append(os.path.sep.join(parts[:i+1]))
        paths.append(os.path.sep.join(parts[:i+1]) + '/config')
        paths.append(os.path.sep.join(parts[:i+1]) + '/.config')

    if any((path.startswith(str(homedir)) for path in paths)):
        paths = [path for path in paths if path.startswith(str(homedir))]

    return paths


class UrsaCtl(App):
    """Ursa Frontier Control primary application."""

    class Meta:
        label = 'ursactl'

        # configuration defaults
        config_defaults = CONFIG

        # call sys.exit() on close
        exit_on_close = True

        # load additional framework extensions
        extensions = [
            'configparser',
            'colorlog',
            'tabulate'
        ]

        hooks = [
            ('post_setup', extend_cache)
        ]

        # configuration handler
        config_handler = 'configparser'

        # configuration file suffix
        config_file_suffix = '.conf'

        # configuration directories
        config_dirs = exploded_paths()

        # set the log handler
        log_handler = 'colorlog'

        # set the output handler
        output_handler = 'tabulate'

        # register handlers
        handlers = [
            Base,
            Update,
            Sync,
            Stop,
            Show,
            Send,
            Run,
            Refresh,
            List,
            Init,
            Get,
            Delete,
            Create
        ]


class UrsaCtlTest(TestApp, UrsaCtl):
    """A sub-class of UrsaCtl that is better suited for testing."""

    class Meta:
        label = 'ursactl'

    def rebuild_cache(self):
        delattr(self, 'db')
        extend_cache(self)


def main():
    with UrsaCtl() as app:
        try:
            app.run()

        except AssertionError as e:
            print('AssertionError > %s' % e.args[0])
            app.exit_code = 1

            if app.debug is True:
                import traceback
                traceback.print_exc()

        except UrsaCtlError as e:
            print('UrsaCtlError > %s' % e.args[0])
            app.exit_code = 1

            if app.debug is True:
                import traceback
                traceback.print_exc()

        except CaughtSignal as e:
            # Default Cement signals are SIGINT and SIGTERM, exit 0 (non-error)
            print('\n%s' % e)
            app.exit_code = 0

        except ConnectionError:
            print('Unable to connect to server')
            app.exit_code = 1


if __name__ == '__main__':
    main()
